# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2015 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

from os import path
import numpy as np
import scipy.optimize as opt
try:
    # scipy's factorial was moved to special in scipy 1.3.0+
    from scipy.special import factorial
except ImportError:
    from scipy.misc import factorial

import warnings

from . import equation_of_state as eos
from .. import constants as constants
from ..tools.chemistry import read_masses
from ..tools.math import bracket

atomic_masses=read_masses()
# energy_states should provide the energies and degeneracies of each electronic level in a variety of elements


class DKS_L(eos.EquationOfState):
    """
    Base class for the finite strain liquid equation of state detailed
    in :cite:`deKoker2013` (supplementary materials).
    """

    """
    Ideal gas contributions (translational and electronic)
    to thermodynamic properties
    """

    def _ln_partition_function(self, mass, temperature):
        """
        Calculates the natural log of the partition function
        """
        return 3./2.*np.log(temperature) \
            + 3./2.*np.log(mass*constants.Boltzmann \
                           /(2*np.pi*constants.Dirac*constants.Dirac)) \

    def _F_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the helmholtz free energy
        Eq. S6, see also eq. 16.72 of Callen., 1985; p. 373
        """

        V = volume/constants.Avogadro
        figoverRT=0.
        for element, N in params['formula'].items(): # N is a.p.f.u
            if N > 1.e-5:
                mass = atomic_masses[element]/constants.Avogadro
                figoverRT += -N*(np.log(V) + self._ln_partition_function(mass, temperature) \
                                     + 1.) + N*np.log(N)
        return constants.gas_constant*temperature*figoverRT


    def _S_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the entropy
        """

        V = volume/constants.Avogadro
        entropy_sum=0.
        for element, N in params['formula'].items(): # N is a.p.f.u
            if N > 1.e-5:
                mass = atomic_masses[element]/constants.Avogadro
                entropy_sum -= -N*(np.log(V) + self._ln_partition_function(mass, temperature) \
                                     + 5./2.) + N*np.log(N)
        return constants.gas_constant*entropy_sum

    def _C_v_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the heat capacity
        """

        n_atoms=0
        for element, N in params['formula'].items():
            n_atoms += N
        return 1.5*constants.gas_constant*n_atoms

    def _P_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the pressure
        PV = nRT
        """

        n_atoms=0
        for element, N in params['formula'].items():
            n_atoms += N
        return n_atoms*constants.gas_constant*temperature / volume

    def _K_T_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the isothermal bulk modulus
        V * d/dV(-nRT/V) = V*nRT/V^2
        """
        n_atoms=0
        for element, N in params['formula'].items():
            n_atoms += N
        return n_atoms*constants.gas_constant*temperature / volume

    def _alphaK_T_ig(self, temperature, volume, params):
        """
        The ideal gas contribution to the product of the
        thermal expansivity and isothermal bulk modulus
        d/dT(nRT/V) = nR/V
        """

        n_atoms=0
        for element, N in params['formula'].items():
            n_atoms += N
        return n_atoms*constants.gas_constant / volume

    """
    Electronic contributions to thermodynamic properties
    """

    def _zeta(self, temperature, volume, params): # eq. S5a, beta in deKoker thesis (3.34)
        return params['zeta_0']*(np.power(volume/params['el_V_0'], params['xi']))

    def _dzetadV(self, temperature, volume, params):
        return params['zeta_0']*params['xi']*(np.power(volume/params['el_V_0'], params['xi']))/volume

    def _d2zetadV2(self, temperature, volume, params):
        return params['zeta_0'] \
            * params['xi'] * (params['xi'] - 1.) \
            * (np.power(volume/params['el_V_0'], params['xi'])) \
            / volume / volume

    def _Tel(self, temperature, volume, params): # eq. S5b
        return params['Tel_0']*(np.power(volume/params['el_V_0'], params['eta']))

    def _dTeldV(self, temperature, volume, params):
        return params['Tel_0'] * params['eta'] \
            * (np.power(volume/params['el_V_0'], params['eta'])) \
            / volume

    def _d2TeldV2(self, temperature, volume, params):
        return params['Tel_0'] \
            * params['eta'] * (params['eta'] - 1.) \
            * (np.power(volume/params['el_V_0'], params['eta'])) \
            / volume / volume

    def _gimel(self, temperature_el, temperature, volume, params): # -F_el/zeta, 3.30 in de Koker thesis
        return 0.5*(temperature*temperature - temperature_el*temperature_el) \
            - temperature*temperature_el*np.log(temperature/temperature_el)

    def _dgimeldTel(self, temperature_el, temperature, volume, params):
        return (temperature-temperature_el) - temperature*np.log(temperature/temperature_el)

    def _dgimeldT(self, temperature_el, temperature, volume, params):
        return (temperature-temperature_el) - temperature_el*np.log(temperature/temperature_el)

    def _d2gimeldTdTel(self, temperature_el, temperature, volume, params):
        return -np.log(temperature/temperature_el)

    def _d2gimeldTel2(self, temperature_el, temperature, volume, params):
        return (temperature/temperature_el)  - 1.

    def _F_el(self, temperature, volume, params): # F_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            F_el = 0
        else:
            F_el = -self._zeta(temperature, volume, params) \
                * self._gimel(temperature_el, temperature, volume, params)
        return F_el

    def _S_el(self, temperature, volume, params): # S_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            S_el = 0
        else:
            S_el = self._zeta(temperature, volume, params) \
                * self._dgimeldT(temperature_el, temperature, volume, params)
        return S_el


    def _P_el(self, temperature, volume, params): # P_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            P_el = 0
        else:
            P_el =  self._dzetadV(temperature, volume, params) \
                * self._gimel(temperature_el, temperature, volume, params) \
                + self._zeta(temperature, volume, params) \
                * self._dTeldV(temperature, volume, params) \
                * self._dgimeldTel(temperature_el, temperature, volume, params)
        return P_el

    def _K_T_el(self, temperature, volume, params): # K_T_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            K_T_el = 0
        else:
            K_T_el =  -volume \
                * ( self._d2zetadV2(temperature, volume, params) \
                        * self._gimel(temperature_el, temperature, volume, params) \
                        + 2. * self._dzetadV(temperature, volume, params) \
                        * self._dgimeldTel(temperature_el, temperature, volume, params) \
                        * self._dTeldV(temperature, volume, params) \
                        + self._zeta(temperature, volume, params) \
                        * ( self._d2TeldV2(temperature, volume, params) \
                                * self._dgimeldTel(temperature_el, temperature, volume, params) \
                                + self._dTeldV(temperature, volume, params) \
                                * self._dTeldV(temperature, volume, params) \
                                * self._d2gimeldTel2(temperature_el, temperature, volume, params)))
        return K_T_el

    def _alphaK_T_el(self, temperature, volume, params): # (alphaK_T)_el
        temperature_el = self._Tel(temperature, volume, params)
        if temperature < temperature_el:
            alphaK_T_el = 0
        else:
            alphaK_T_el = self._dzetadV(temperature, volume, params) \
                * self._dgimeldT(temperature_el, temperature, volume, params) \
                + self._zeta(temperature, volume, params) \
                * self._d2gimeldTdTel(temperature_el, temperature, volume, params) \
                * self._dTeldV(temperature, volume, params)
        return alphaK_T_el

    def _C_v_el(self, temperature, volume, params): # C_el, eq. 3.28 of de Koker thesis
        temperature_el = self._Tel(temperature, volume, params)
        zeta = self._zeta(temperature, volume, params)

        if temperature > temperature_el:
            Cv_el = zeta*(temperature - temperature_el)
        else:
            Cv_el = 0.
        return Cv_el



    """
    Excess (bonding) contributions to thermodynamic properties
    """

    # Finite strain
    def _finite_strain(self, temperature, volume, params): # f(V), eq. S3a
        return (1./2.)*(np.power(params['V_0']/volume, 2./3.) - 1.0)

    def _dfdV(self, temperature, volume, params): # f(V), eq. S3a
        return (-1./3.)*np.power(params['V_0']/volume, 2./3.)/volume

    def _d2fdV2(self,temperature, volume, params):
        return (5./9.)*np.power(params['V_0']/volume, 2./3.)/volume/volume

    # Temperature
    def _theta(self, temperature, volume, params): # theta, eq. S3b
        return np.power(temperature/params['T_0'], params['m']) - 1.

    def _dthetadT(self, temperature, volume, params):
        return params['m']*np.power(temperature/params['T_0'], params['m']) \
            / temperature

    def _d2thetadT2(self, temperature, volume, params):
        return params['m']*(params['m']-1.)*np.power(temperature/params['T_0'], params['m']) \
            / temperature / temperature

    def _F_xs(self, temperature, volume, params): # F_xs, eq. S2
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        energy = 0.
        for i in range(len(params['a'])):
            ifact=factorial(i, exact=False)
            for j in range(len(params['a'][0])):
                jfact=factorial(j, exact=False)
                energy += params['a'][i][j]*np.power(f, i)*np.power(theta, j)/ifact/jfact
        return energy

    def _S_xs(self, temperature, volume, params): # F_xs, eq. 3.18
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        entropy = 0.
        for i in range(len(params['a'])):
            ifact = factorial(i, exact=False)
            for j in range(len(params['a'][0])):
                if j > 0:
                    jfact = factorial(j, exact=False)
                    entropy += j*params['a'][i][j]*np.power(f, i)*np.power(theta, j-1.)/ifact/jfact
        return -self._dthetadT(temperature, volume, params)*entropy

    def _P_xs(self, temperature, volume, params): # P_xs, eq. 3.17 of de Koker thesis
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        pressure=0.
        for i in range(len(params['a'])):
            ifact=factorial(i, exact=False)
            if i > 0:
                for j in range(len(params['a'][0])):
                    jfact=factorial(j, exact=False)
                    pressure += float(i)*params['a'][i][j]*np.power(f, float(i)-1.)*np.power(theta, float(j))/ifact/jfact
        return -self._dfdV(temperature, volume, params)*pressure

    def _K_T_xs(self, temperature, volume, params): # K_T_xs, eq. 3.20 of de Koker thesis
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        K_ToverV=0.
        for i in range(len(params['a'])):
            ifact=factorial(i, exact=False)
            for j in range(len(params['a'][0])):
                if i > 0:
                    jfact=factorial(j, exact=False)
                    prefactor = float(i) * params['a'][i][j] \
                        * np.power(theta, float(j)) / ifact / jfact
                    K_ToverV += prefactor*self._d2fdV2(temperature, volume, params) \
                        * np.power(f, float(i-1))
                if i > 1:
                    dfdV = self._dfdV(temperature, volume, params)
                    K_ToverV += prefactor * dfdV * dfdV \
                        * float(i-1) * np.power(f, float(i-2))
        return volume*K_ToverV

    def _alphaK_T_xs(self, temperature, volume, params): # eq. 3.21 of de Koker thesis
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        sum_factors = 0.
        for i in range(len(params['a'])):
            ifact=factorial(i, exact=False)
            if i > 0:
                for j in range(len(params['a'][0])):
                    if j > 0:
                        jfact=factorial(j, exact=False)
                        sum_factors += float(i)*float(j)*params['a'][i][j] \
                            * np.power(f, float(i-1)) * np.power(theta, float(j-1)) \
                            / ifact / jfact

        return -self._dfdV(temperature, volume, params) \
            * self._dthetadT(temperature, volume, params) \
            * sum_factors


    def _C_v_xs(self, temperature, volume, params): # Cv_xs, eq. 3.22 of de Koker thesis
        f = self._finite_strain(temperature, volume, params)
        theta = self._theta(temperature, volume, params)
        C_voverT=0.
        for i in range(len(params['a'])):
            ifact=factorial(i, exact=False)
            for j in range(len(params['a'][0])):
                if j > 0:
                    jfact=factorial(j, exact=False)
                    prefactor = float(j)*params['a'][i][j]*np.power(f, float(i))/ifact/jfact
                    C_voverT += prefactor * self._d2thetadT2(temperature, volume, params) \
                        * np.power(theta, float(j-1))
                if j > 1:
                    dthetadT = self._dthetadT(temperature, volume, params)
                    C_voverT += prefactor * dthetadT * dthetadT \
                        * float(j-1) * np.power(theta, float(j-2))
        return -temperature*C_voverT


    """
    Magnetic contributions to thermodynamic properties
    (as found in Ramo and Stixrude, 2014)
    """

    def _spin(self, temperature, volume, params):
        S_a = 0.
        S_b = 0.
        numerator = 0.
        numerator_2 = 0.
        n_atoms = 0.
        if 'spin_a' in params:
            for element, N in params['formula'].items():
                if element == 'Fe':
                    n_atoms += N

            VoverVx = volume/params['V_0']
            S_a = params['spin_a'][0] + params['spin_a'][1]*VoverVx
            S_b = (params['spin_b'][0]
                   + params['spin_b'][1]/VoverVx
                   + params['spin_b'][2]/(np.power(VoverVx, 2.))
                   + params['spin_b'][3]/(np.power(VoverVx, 3.)))

            # S = S_a*T + S_b
            # d(2S + 1)/dV
            numerator=-2.*(-params['spin_a'][1]*temperature
                           + params['spin_b'][1]/(np.power(VoverVx, 2.))
                           + 2.*params['spin_b'][2]/(np.power(VoverVx, 3.))
                           + 3.*params['spin_b'][3]/(np.power(VoverVx, 4.)))/params['V_0']

            # d2S/dV2
            numerator_2 = 2.*((2.*params['spin_b'][1]/(np.power(VoverVx, 3.))
                               + 6.*params['spin_b'][2]/(np.power(VoverVx, 4.))
                               + 12.*params['spin_b'][3]/(np.power(VoverVx, 5.)))
                              /np.power(params['V_0'], 2.))
        return S_a, S_b, numerator, numerator_2, n_atoms


    def _F_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(temperature, volume, params)
        S = S_a*temperature + S_b
        return -n_atoms*constants.gas_constant*temperature*np.log(2.*S + 1.)


    def _S_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(temperature, volume, params)
        S = S_a*temperature + S_b
        return n_atoms*constants.gas_constant * ((2.*S_a*temperature/(2.*S + 1.)
                                                   + np.log(2.*S + 1.)))

    def _P_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(temperature, volume, params)
        S = S_a*temperature + S_b
        dFdV = -n_atoms*constants.gas_constant*temperature*numerator/(2.*S + 1.)
        return -dFdV


    def _K_T_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(temperature, volume, params)
        S = S_a*temperature + S_b
        dFdV = numerator/(2.*S + 1.)
        d2FdV2 = numerator_2/(2.*S + 1.) - np.power(dFdV, 2.)

        return -volume*n_atoms*constants.gas_constant*temperature*d2FdV2

    def _alphaK_T_mag(self, temperature, volume, params): # WARNING: numeric differentiation a.t.m.
        return (self._P_mag(temperature + 0.5, volume, params)
                - self._P_mag(temperature - 0.5, volume, params))


    def _C_v_mag(self, temperature, volume, params):
        S_a, S_b, numerator, numerator_2, n_atoms = self._spin(temperature, volume, params)
        S = S_a*temperature + S_b
        return n_atoms * constants.gas_constant * temperature * 4.*S_a*(S_a*temperature + 2.*S_b + 1.)/np.power(2.*S + 1., 2.)


    def _aK_T(self, temperature, volume, params):
        aK_T =  (self._alphaK_T_ig(temperature, volume, params)
                 + self._alphaK_T_el(temperature, volume, params)
                 + self._alphaK_T_xs(temperature, volume, params)
                 + self._alphaK_T_mag(temperature, volume, params))
        return aK_T

    # Pressure
    def pressure(self, temperature, volume, params):
        P = (self._P_ig(temperature, volume, params)
             + self._P_el(temperature, volume, params)
             + self._P_xs(temperature, volume, params)
             + self._P_mag(temperature, volume, params))
        return P

    def volume(self, pressure, temperature, params):
        _delta_pressure = lambda x, pressure, temperature, params: pressure - self.pressure(temperature, x, params)

        # we need to have a sign change in [a,b] to find a zero. Let us start with a
        # conservative guess:
        args = (pressure, temperature, params)
        try:
            sol = bracket(_delta_pressure, params['V_0'],
                          1.e-2 * params['V_0'], args)
        except ValueError:
            raise Exception(
                'Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?')
        return opt.brentq(_delta_pressure, sol[0], sol[1], args=args)

    def isothermal_bulk_modulus(self, pressure,temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`[Pa]`
        """
        K_T = (self._K_T_ig(temperature, volume, params)
               + self._K_T_el(temperature, volume, params)
               + self._K_T_xs(temperature, volume, params)
               + self._K_T_mag(temperature, volume, params))
        return K_T

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus. :math:`[Pa]`
        """
        K_S = (self.isothermal_bulk_modulus(pressure,temperature, volume, params)
               * ( 1. + temperature
                   * self.thermal_expansivity(pressure, temperature, volume, params)
                   * self.grueneisen_parameter(pressure, temperature, volume, params)))
        return K_S

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter. :math:`[unitless]`
        """
        gamma = (self._aK_T(temperature, volume, params)
                 * volume
                 / self.molar_heat_capacity_v(pressure, temperature, volume, params))
        return gamma

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus. :math:`[Pa]`
        Zero for fluids
        """
        return 0.

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume. :math:`[J/K/mol]`
        """
        C_v = (self._C_v_ig(temperature, volume, params)
               + self._C_v_el(temperature, volume, params)
               + self._C_v_xs(temperature, volume, params)
               + self._C_v_mag(temperature, volume, params))
        return C_v

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure. :math:`[J/K/mol]`
        """
        C_p = (self.molar_heat_capacity_v(pressure,temperature, volume, params)
               * ( 1. + temperature
                   * self.thermal_expansivity(pressure, temperature, volume, params)
                   * self.grueneisen_parameter(pressure, temperature, volume, params) ))
        return C_p

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity. :math:`[1/K]`
        """
        alpha = (self._aK_T(temperature, volume, params)
                 / self.isothermal_bulk_modulus(0., temperature, volume, params))
        return alpha

    def gibbs_free_energy( self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy at the pressure and temperature of the mineral [J/mol]
        """
        G = self.helmholtz_free_energy( pressure, temperature, volume, params) + pressure * volume
        return G

    def entropy( self, pressure, temperature, volume, params):
        """
        Returns the entropy at the pressure and temperature of the mineral [J/K/mol]
        """
        S = (self._S_ig(temperature, volume, params)
             + self._S_el(temperature, volume, params)
             + self._S_xs(temperature, volume, params)
             + self._S_mag(temperature, volume, params))
        return S

    def enthalpy( self, pressure, temperature, volume, params):
        """
        Returns the enthalpy at the pressure and temperature of the mineral [J/mol]
        """
        H = self.helmholtz_free_energy( pressure, temperature, volume, params) + \
            temperature * self.entropy( pressure, temperature, volume, params) + \
            pressure * self.volume( pressure, temperature, params)
        return H

    def helmholtz_free_energy( self, pressure, temperature, volume, params):
        """
        Returns the Helmholtz free energy at the pressure and temperature of the mineral [J/mol]
        """
        F = (self._F_ig(temperature, volume, params)
             + self._F_el(temperature, volume, params)
             + self._F_xs(temperature, volume, params)
             + self._F_mag(temperature, volume, params))
        return F

    def molar_internal_energy(self, pressure, temperature, volume, params):
        E = self.helmholtz_free_energy(pressure, temperature, volume, params) + \
            temperature*self.entropy(pressure, temperature, volume, params)
        return E

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        # Check that all the required keys are in the dictionary
        expected_keys = ['V_0', 'T_0', 'O_theta', 'O_f', 'm', 'a', 'zeta_0', 'xi', 'Tel_0', 'eta']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Sometimes the standard electronic volume is different to V_0.
        # If not, make it the same.
        if 'el_V_0' not in params:
            params['el_V_0'] = params['V_0']



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import matplotlib.pyplot as plt


def pretty_plot():
    """
    Makes pretty plots. Overwrites the matplotlib default settings to allow
    for better fonts. Slows down plotting
    """
    import matplotlib.pyplot as plt
    plt.rc('text', usetex=True)
    plt.rcParams['text.latex.preamble'] = '\\usepackage{relsize}'
    plt.rc('font', family='sanserif')


def plot_projected_elastic_properties(mineral, plot_types, axes,
                                      n_zenith=31, n_azimuth=91,
                                      n_divs=100):
    """
    Plot types must be one of:
    'vp': V$_{P}$ (km/s)
    'vs1': 'V$_{S1}$ (km/s)
    'vs2': V$_{S2}$ (km/s)
    'vp/vs1': V$_{P}$/V$_{S1}$
    'vp/vs2': V$_{P}$/V$_{S2}$
    's anisotropy': S-wave anisotropy (%), 200(vs1s - vs2s)/(vs1s + vs2s))
    'linear compressibility' Linear compressibility (GPa$^{-1}$)
    'youngs modulus': Youngs Modulus (GPa)

    axes objects must be initialised with projection='polar'
    """

    assert len(plot_types) == len(axes)

    zeniths = np.linspace(np.pi/2., np.pi, n_zenith)
    azimuths = np.linspace(0., 2.*np.pi, n_azimuth)
    Rs = np.sin(zeniths)/(1. - np.cos(zeniths))
    r, theta = np.meshgrid(Rs, azimuths)

    vps = np.empty_like(r)
    vs1s = np.empty_like(r)
    vs2s = np.empty_like(r)
    betas = np.empty_like(r)
    Es = np.empty_like(r)
    for i, az in enumerate(azimuths):
        for j, phi in enumerate(zeniths):
            d = np.array([np.cos(az)*np.sin(phi), np.sin(az)*np.sin(phi),
                          -np.cos(phi)])  # change_hemispheres
            velocities = mineral.wave_velocities(d)
            betas[i][j] = mineral.isentropic_linear_compressibility(d)
            Es[i][j] = mineral.isentropic_youngs_modulus(d)
            vps[i][j] = velocities[0][0]
            vs1s[i][j] = velocities[0][1]
            vs2s[i][j] = velocities[0][2]

    prps = []
    for type in plot_types:
        if type == 'vp':
            prps.append(('V$_{P}$ (km/s)', vps/1000.))
        elif type == 'vs1':
            prps.append(('V$_{S1}$ (km/s)', vs1s/1000.))
        elif type == 'vs2':
            prps.append(('V$_{S2}$ (km/s)', vs2s/1000.))
        elif type == 'vp/vs1':
            prps.append(('V$_{P}$/V$_{S1}$', vps/vs1s))
        elif type == 'vp/vs2':
            prps.append(('V$_{P}$/V$_{S2}$', vps/vs2s))
        elif type == 's anisotropy':
            prps.append(('S-wave anisotropy (%)',
                         200.*(vs1s - vs2s)/(vs1s + vs2s)))
        elif type == 'linear compressibility':
            prps.append(('Linear compressibility (GPa$^{-1}$)', betas*1.e9))
        elif type == 'youngs modulus':
            prps.append(('Youngs Modulus (GPa)', Es/1.e9))
        else:
            raise Exception('plot_type not recognised.')

    contour_sets = []
    ticks = []
    lines = []
    for i, prp in enumerate(prps):
        title, item = prp

        axes[i].set_title(title)

        vmin = np.min(item)
        vmax = np.max(item)
        spacing = np.power(10., np.floor(np.log10(vmax - vmin)))
        nt = int((vmax - vmin - vmax % spacing + vmin % spacing) / spacing)
        if nt == 1:
            spacing = spacing/4.
        elif nt < 4:
            spacing = spacing/2.
        elif nt > 8:
            spacing = spacing*2.

        tmin = vmin + (spacing - vmin % spacing)
        tmax = vmax - vmax % spacing
        nt = int((tmax - tmin)/spacing + 1)

        ticks.append(np.linspace(tmin, tmax, nt))
        contour_sets.append(axes[i].contourf(theta, r, item, n_divs,
                                             cmap=plt.cm.jet_r,
                                             vmin=vmin, vmax=vmax))
        lines.append(axes[i].contour(theta, r, item, ticks[-1],
                                     colors=('black',), linewidths=(1,)))
        axes[i].set_yticks([100])

    return contour_sets, ticks, lines



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import
from __future__ import print_function
import warnings

import numpy as np

from .material import Material, material_property
from .. import eos
from ..tools.misc import copy_documentation


class Mineral(Material):

    """
    This is the base class for all minerals. States of the mineral
    can only be queried after setting the pressure and temperature
    using set_state(). The method for computing properties of
    the material is set using set_method(). This is done during
    initialisation if the param 'equation_of_state' has been defined.
    The method can be overridden later by the user.

    This class is available as ``burnman.Mineral``.

    If deriving from this class, set the properties in self.params
    to the desired values. For more complicated materials you
    can overwrite set_state(), change the params and then call
    set_state() from this class.

    All the material parameters are expected to be in plain SI units.  This
    means that the elastic moduli should be in Pascals and NOT Gigapascals,
    and the Debye temperature should be in K not C.  Additionally, the
    reference volume should be in m^3/(mol molecule) and not in unit cell
    volume and 'n' should be the number of atoms per molecule.  Frequently in
    the literature the reference volume is given in Angstrom^3 per unit cell.
    To convert this to m^3/(mol of molecule) you should multiply by 10^(-30) *
    N_a / Z, where N_a is Avogadro's number and Z is the number of formula units per
    unit cell. You can look up Z in many places, including www.mindat.org
    """

    def __init__(self, params=None, property_modifiers=None):
        Material.__init__(self)
        if params is not None:
            self.params = params
        elif 'params' not in self.__dict__:
            self.params = {}

        if property_modifiers is not None:
            self.property_modifiers = property_modifiers
        elif 'property_modifiers' not in self.__dict__:
            self.property_modifiers = []

        self.method = None
        if 'equation_of_state' in self.params:
            self.set_method(self.params['equation_of_state'])
        if 'name' in self.params:
            self.name = self.params['name']

    def set_method(self, equation_of_state):
        """
        Set the equation of state to be used for this mineral.
        Takes a string corresponding to any of the predefined
        equations of state:  'bm2', 'bm3', 'mgd2', 'mgd3', 'slb2', 'slb3',
        'mt', 'hp_tmt', or 'cork'.  Alternatively, you can pass a user defined
        class which derives from the equation_of_state base class.
        After calling set_method(), any existing derived properties
        (e.g., elastic parameters or thermodynamic potentials) will be out
        of date, so set_state() will need to be called again.
        """

        if equation_of_state is None:
            self.method = None
            return

        new_method = eos.create(equation_of_state)
        if self.method is not None and 'equation_of_state' in self.params:
            self.method = eos.create(self.params['equation_of_state'])

        if type(new_method).__name__ == 'instance':
            raise Exception(
                "Please derive your method from object (see python old style classes)")

        if ((self.method is not None
             and isinstance(new_method, type(self.method)) is False)):

            # Warn user that they are changing the EoS
            warnings.warn('Warning, you are changing the method to '
                          f'{new_method.__class__.__name__} even though the '
                          'material is designed to be used with the method '
                          f'{self.method.__class__.__name__}. '
                          'This does not overwrite any mineral attributes',
                          stacklevel=2)
            self.reset()

        self.method = new_method

        # Validate the params object on the requested EOS.
        try:
            self.method.validate_parameters(self.params)
        except Exception as e:
            print(f'Mineral {self.to_string()} failed to validate parameters '
                  f'with message: \"{e.message}\"')
            raise

        # Invalidate the cache upon resetting the method
        self.reset()

    def to_string(self):
        """
        Returns the name of the mineral class
        """
        return "'" + self.__class__.__module__.replace(".minlib_", ".") + "." + self.__class__.__name__ + "'"

    def debug_print(self, indent=""):
        print("%s%s" % (indent, self.to_string()))

    def unroll(self):
        return ([self], [1.0])

    @copy_documentation(Material.set_state)
    def set_state(self, pressure, temperature):
        Material.set_state(self, pressure, temperature)
        self._property_modifiers = eos.property_modifiers.calculate_property_modifications(
            self)

        if self.method is None:
            raise AttributeError(
                "no method set for mineral, or equation_of_state given in mineral.params")

    """
    Properties from equations of state
    We choose the P, T properties (e.g. Gibbs(P, T) rather than Helmholtz(V, T)),
    as it allows us to more easily apply corrections to the free energy
    """
    @material_property
    @copy_documentation(Material.molar_gibbs)
    def molar_gibbs(self):
        return self.method.gibbs_free_energy(self.pressure, self.temperature, self._molar_volume_unmodified, self.params) \
            + self._property_modifiers['G']

    @material_property
    def _molar_volume_unmodified(self):
        return self.method.volume(self.pressure, self.temperature, self.params)

    @material_property
    @copy_documentation(Material.molar_volume)
    def molar_volume(self):
        return self._molar_volume_unmodified \
            + self._property_modifiers['dGdP']

    @material_property
    @copy_documentation(Material.molar_entropy)
    def molar_entropy(self):
        return self.method.entropy(self.pressure, self.temperature, self._molar_volume_unmodified, self.params) \
            - self._property_modifiers['dGdT']

    @material_property
    @copy_documentation(Material.isothermal_bulk_modulus)
    def isothermal_bulk_modulus(self):
        K_T_orig = self.method.isothermal_bulk_modulus(
            self.pressure, self.temperature,
            self._molar_volume_unmodified, self.params)

        return self.molar_volume \
            / ((self._molar_volume_unmodified / K_T_orig) - self._property_modifiers['d2GdP2'])

    @material_property
    @copy_documentation(Material.molar_heat_capacity_p)
    def molar_heat_capacity_p(self):
        return (self.method.molar_heat_capacity_p(self.pressure,
                                                  self.temperature,
                                                  self._molar_volume_unmodified,
                                                  self.params)
                - self.temperature * self._property_modifiers['d2GdT2'])

    @material_property
    @copy_documentation(Material.thermal_expansivity)
    def thermal_expansivity(self):
        return (
            (self.method.thermal_expansivity(self.pressure, self.temperature,
                                             self._molar_volume_unmodified,
                                             self.params)
             * self._molar_volume_unmodified)
            + self._property_modifiers['d2GdPdT']) / self.molar_volume

    @material_property
    @copy_documentation(Material.shear_modulus)
    def shear_modulus(self):
        G = self.method.shear_modulus(
            self.pressure, self.temperature, self._molar_volume_unmodified,
            self.params)
        if G < np.finfo('float').eps:
            warnings.formatwarning = lambda msg, * \
                a: 'Warning from file \'{0}\', line {1}:\n{2}\n\n'.format(
                    a[1], a[2], msg)
            warnings.warn('You are trying to calculate shear modulus for {0} when it is exactly zero. \n'
                          'If {0} is a liquid, then you can safely ignore this warning, but consider \n'
                          'calculating bulk modulus or bulk sound rather than Vp or Vs. \n'
                          'If {0} is not a liquid, then shear modulus calculations for the \n'
                          'underlying equation of state ({1}) have not been implemented, \n'
                          'and Vp and Vs estimates will be incorrect.'.format(self.name, self.method.__class__.__name__), stacklevel=1)
        return G

    """
    Properties from mineral parameters,
    Legendre transformations
    or Maxwell relations
    """
    @material_property
    def formula(self):
        """
        Returns the chemical formula of the Mineral class
        """
        if 'formula' in self.params:
            return self.params['formula']
        else:
            raise ValueError(
                'No formula parameter for mineral {0}.'.format(self.to_string))

    @material_property
    @copy_documentation(Material.molar_mass)
    def molar_mass(self):
        if 'molar_mass' in self.params:
            return self.params['molar_mass']
        else:
            raise ValueError(
                'No molar_mass parameter for mineral {0}.'.format(self.to_string))

    @material_property
    @copy_documentation(Material.density)
    def density(self):
        return self.molar_mass / self.molar_volume

    @material_property
    @copy_documentation(Material.molar_internal_energy)
    def molar_internal_energy(self):
        return self.molar_gibbs - self.pressure * self.molar_volume + self.temperature * self.molar_entropy

    @material_property
    @copy_documentation(Material.molar_helmholtz)
    def molar_helmholtz(self):
        return self.molar_gibbs - self.pressure * self.molar_volume

    @material_property
    @copy_documentation(Material.molar_enthalpy)
    def molar_enthalpy(self):
        return self.molar_gibbs + self.temperature * self.molar_entropy

    @material_property
    @copy_documentation(Material.adiabatic_bulk_modulus)
    def adiabatic_bulk_modulus(self):
        if self.temperature < 1.e-10:
            return self.isothermal_bulk_modulus
        else:
            return self.isothermal_bulk_modulus * self.molar_heat_capacity_p / self.molar_heat_capacity_v

    @material_property
    @copy_documentation(Material.isothermal_compressibility)
    def isothermal_compressibility(self):
        return 1. / self.isothermal_bulk_modulus

    @material_property
    @copy_documentation(Material.adiabatic_compressibility)
    def adiabatic_compressibility(self):
        return 1. / self.adiabatic_bulk_modulus

    @material_property
    @copy_documentation(Material.p_wave_velocity)
    def p_wave_velocity(self):
        return np.sqrt((self.adiabatic_bulk_modulus
                        + 4. / 3. * self.shear_modulus) / self.density)

    @material_property
    @copy_documentation(Material.bulk_sound_velocity)
    def bulk_sound_velocity(self):
        return np.sqrt(self.adiabatic_bulk_modulus / self.density)

    @material_property
    @copy_documentation(Material.shear_wave_velocity)
    def shear_wave_velocity(self):
        return np.sqrt(self.shear_modulus / self.density)

    @material_property
    @copy_documentation(Material.grueneisen_parameter)
    def grueneisen_parameter(self):
        eps = np.finfo('float').eps
        if np.abs(self.molar_heat_capacity_v) > eps:

            return (self.thermal_expansivity
                    * self.isothermal_bulk_modulus
                    * self.molar_volume
                    / self.molar_heat_capacity_v)
        elif ((np.abs(self._property_modifiers['d2GdPdT']) < eps)
              and (np.abs(self._property_modifiers['d2GdP2']) < eps)
              and (np.abs(self._property_modifiers['dGdP']) < eps)
              and (np.abs(self._property_modifiers['d2GdT2']) < eps)):

            return self.method.grueneisen_parameter(self.pressure, self.temperature,
                                                    self.molar_volume, self.params)
        else:
            raise Exception(
                'You are trying to calculate the grueneisen parameter at a temperature where the heat capacity is very low and where you have defined Gibbs property modifiers.')

    @material_property
    @copy_documentation(Material.molar_heat_capacity_v)
    def molar_heat_capacity_v(self):
        return self.molar_heat_capacity_p - self.molar_volume * self.temperature \
            * self.thermal_expansivity * self.thermal_expansivity \
            * self.isothermal_bulk_modulus



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Irving et al. 2018
^^^^^^^^^^^^^

Seismically derived EoS parameters for the outer core using normal mode center frequencies.
These parameters are derived parameterizing the outer core velocity and density with an equation-of-state, and inverting for K, Kprime, and molar_volume (molar_mass is fixed to 0.05, as sensitivy of the normal modes is only to molar density). The equations-of-state are assumed to capture the isentropic behaviour across the outer core. Potentially, these parameters can be used to represent cores of other planets.
See:
Irving, Cottaar, Lekic (2018) Seismically determined elastic parameters for the outer core, Science Advances

"""
from __future__ import absolute_import

from ..classes import mineral_helpers as helpers
from ..classes.mineral import Mineral


class EPOC_vinet (Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'vinet',
            'V_0': 8.18e-6,
            'K_0': 67.5e9,
            'Kprime_0': 6.12,
            'molar_mass': .05,}
        Mineral.__init__(self)

class EPOC_bm (Mineral):

    def __init__(self):
        self.params = {
            'equation_of_state': 'bm3',
            'V_0': 7.63e-6,
            'K_0': 120e9,
            'Kprime_0': 4.6,
            'molar_mass': .05,}
        Mineral.__init__(self)



from __future__ import print_function
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
from copy import deepcopy


# TODO: When we require Python 3.8+, replace with
# functools.cached_property decorator
class cached_property(property):

    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work.

    This decorator is adapted slightly from the one in the werkzeug module:
    https://tedboy.github.io/flask/_modules/werkzeug/utils.html#cached_property
    """

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = value

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        try:
            value = obj.__dict__[self.__name__]
        except KeyError:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


def material_property(func):
    """
    Decorator @material_property to be used for cached properties of materials.

    To be used on function in Material or derived classes that should be exposed
    as read-only properties that are cached. The function Material.reset() will
    reset the cached values.

    Internally, the values are stored in a dictionary member called _cached, which
    is emptied by .reset().
    """
    class mat_obj():

        def __init__(self, func):
            self.func = func
            self.varname = self.func.__name__

        def get(self, obj):
            if not hasattr(obj, "_cached"):
                raise Exception("The material_property decorator could not find class member _cached. "
                                "Did you forget to call Material.__init__(self) in __init___?")
            cache_array = getattr(obj, "_cached")
            if self.varname not in cache_array:
                cache_array[self.varname] = self.func(obj)
            return cache_array[self.varname]

    return property(mat_obj(func).get, doc=func.__doc__)


class Material(object):

    """
    Base class for all materials. The main functionality is unroll() which
    returns a list of objects of type :class:`~burnman.Mineral` and their molar
    fractions. This class is available as ``burnman.Material``.

    The user needs to call set_method() (once in the beginning) and set_state()
    before querying the material with unroll() or density().
    """

    def __init__(self):
        self._pressure = None
        self._temperature = None
        if not hasattr(self, "name"):
            # if a derived class decides to set .name before calling this
            # constructor (I am looking at you, SLB_2011.py!), do not
            # overwrite the name here.
            self._name = self.__class__.__name__
        self._cached = {}

    @property
    def name(self):
        """ Human-readable name of this material.

        By default this will return the name of the class, but it can be set
        to an arbitrary string. Overriden in Mineral.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    def set_method(self, method):
        """
        Set the averaging method. See :doc:`averaging` for details.

        Notes
        -----
        Needs to be implemented in derived classes.
        """
        raise NotImplementedError(
            "need to implement set_method() in derived class!")

    def to_string(self):
        """
        Returns a human-readable name of this material. The default implementation will return the name of the class,
        which is a reasonable default.

        Returns
        -------
        name : string
            Name of this material.
        """
        return "'" + self.name + "'"

    def debug_print(self, indent=""):
        """
        Print a human-readable representation of this Material.
        """
        raise NotImplementedError(
            "Derived classes need to implement debug_print(). This is '" + self.__class__.__name__ + "'")

    def print_minerals_of_current_state(self):
        """
        Print a human-readable representation of this Material at the current P, T as a list of minerals.
        This requires set_state() has been called before.
        """
        (minerals, fractions) = self.unroll()
        if len(minerals) == 1:
            print(minerals[0].to_string())
        else:
            print("Material %s:" % self.to_string())
            for (mineral, fraction) in zip(minerals, fractions):
                print("  %g of phase %s" % (fraction, mineral.to_string()))

    def set_state(self, pressure, temperature):
        """
        Set the material to the given pressure and temperature.

        Parameters
        ----------
        pressure : float
            The desired pressure in [Pa].
        temperature : float
            The desired temperature in [K].
        """
        if not hasattr(self, "_pressure"):
            raise Exception("Material.set_state() could not find class member _pressure. "
                            "Did you forget to call Material.__init__(self) in __init___?")
        self.reset()

        self._pressure = pressure
        self._temperature = temperature

    def reset(self):
        """
        Resets all cached material properties.

        It is typically not required for the user to call this function.
        """
        self._cached = {}

    def copy(self):
        return deepcopy(self)

    def unroll(self):
        """
        Unroll this material into a list of :class:`burnman.Mineral` and their molar fractions. All averaging schemes
        then operate on this list of minerals. Note that the return value of this function may depend on the current
        state (temperature, pressure).

        Notes
        -----
        Needs to be implemented in derived classes.

        Returns
        -------
        fractions : list of float
            List of molar fractions, should sum to 1.0.
        minerals : list of :class:`burnman.Mineral`
            List of minerals.
        """
        raise NotImplementedError(
            "need to implement unroll() in derived class!")

    def evaluate(self, vars_list, pressures, temperatures):
        """
        Returns an array of material properties requested through a list of strings at given pressure and temperature
        conditions. At the end it resets the set_state to the original values.
        The user needs to call set_method() before.

        Parameters
        ----------
        vars_list : list of strings
            Variables to be returned for given conditions
        pressures : ndlist or ndarray of float
            n-dimensional array of pressures in [Pa].
        temperatures : ndlist or ndarray of float
            n-dimensional array of temperatures in [K].

        Returns
        -------
        output : array of array of float
            Array returning all variables at given pressure/temperature values. output[i][j] is property vars_list[j]
            and temperatures[i] and pressures[i].

        """
        old_pressure = self.pressure
        old_temperature = self.temperature
        pressures = np.array(pressures)
        temperatures = np.array(temperatures)

        assert(pressures.shape == temperatures.shape)

        output = np.empty((len(vars_list),) + pressures.shape)
        for i, p in np.ndenumerate(pressures):
            self.set_state(p, temperatures[i])
            for j in range(len(vars_list)):
                output[(j,) + i] = getattr(self, vars_list[j])
        if old_pressure is None or old_temperature is None:
            # do not set_state if old values were None. Just reset to None
            # manually
            self._pressure = self._temperature = None
            self.reset()
        else:
            self.set_state(old_pressure, old_temperature)

        return output

    @property
    def pressure(self):
        """
        Returns current pressure that was set with :func:`~burnman.Material.set_state`.


        Notes
        -----
        - Aliased with :func:`~burnman.Material.P`.

        Returns
        -------
        pressure : float
            Pressure in [Pa].
        """
        return self._pressure

    @property
    def temperature(self):
        """
        Returns current temperature that was set with :func:`~burnman.Material.set_state`.

        Notes
        -----
        - Aliased with :func:`~burnman.Material.T`.

        Returns
        -------
        temperature : float
            Temperature in [K].
        """
        return self._temperature

    @material_property
    def molar_internal_energy(self):
        """
        Returns the molar internal energy of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.energy`.

        Returns
        -------
        molar_internal_energy : float
            The internal energy in [J/mol].
        """
        raise NotImplementedError(
            "need to implement molar_internal_energy() in derived class!")

    @material_property
    def molar_gibbs(self):
        """
        Returns the molar Gibbs free energy of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.gibbs`.

        Returns
        -------
        molar_gibbs : float
            Gibbs free energy in [J/mol].
        """
        raise NotImplementedError(
            "need to implement molar_gibbs() in derived class!")

    @material_property
    def molar_helmholtz(self):
        """
        Returns the molar Helmholtz free energy of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.helmholtz`.

        Returns
        -------
        molar_helmholtz : float
            Helmholtz free energy in [J/mol].
        """
        raise NotImplementedError(
            "need to implement molar_helmholtz() in derived class!")

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.

        Returns
        -------
        molar_mass : float
            Molar mass in [kg/mol].
        """
        raise NotImplementedError(
            "need to implement molar_mass() in derived class!")

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.V`.

        Returns
        -------
        molar_volume : float
            Molar volume in [m^3/mol].
        """
        raise NotImplementedError(
            "need to implement molar_volume() in derived class!")

    @material_property
    def density(self):
        """
        Returns the density of this material.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.rho`.

        Returns
        -------
        density : float
            The density of this material in [kg/m^3].
        """
        raise NotImplementedError(
            "need to implement density() in derived class!")

    @material_property
    def molar_entropy(self):
        """
        Returns molar entropy of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.S`.

        Returns
        -------
        molar_entropy : float
            Entropy in [J/K/mol].
        """
        raise NotImplementedError(
            "need to implement molar_entropy() in derived class!")

    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpy of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.H`.

        Returns
        -------
        molar_enthalpy : float
            Enthalpy in [J/mol].
        """
        raise NotImplementedError(
            "need to implement molar_enthalpy() in derived class!")

    @material_property
    def isothermal_bulk_modulus(self):
        """
        Returns isothermal bulk modulus of the material.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.K_T`.

        Returns
        -------
        isothermal_bulk_modulus : float
            Bulk modulus in [Pa].
        """
        raise NotImplementedError(
            "need to implement isothermal_bulk_moduls() in derived class!")

    @material_property
    def adiabatic_bulk_modulus(self):
        """
        Returns the adiabatic bulk modulus of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.K_S`.

        Returns
        -------
        adiabatic_bulk_modulus : float
            Adiabatic bulk modulus in [Pa].
        """
        raise NotImplementedError(
            "need to implement adiabatic_bulk_modulus() in derived class!")

    @material_property
    def isothermal_compressibility(self):
        """
        Returns isothermal compressibility of the mineral (or inverse isothermal bulk modulus).

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.beta_T`.

        Returns
        -------
        (K_T)^-1 : float
            Compressibility in [1/Pa].
        """
        raise NotImplementedError(
            "need to implement compressibility() in derived class!")

    @material_property
    def adiabatic_compressibility(self):
        """
        Returns adiabatic compressibility of the mineral (or inverse adiabatic bulk modulus).


        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.beta_S`.

        Returns
        -------
        adiabatic_compressibility : float
            adiabatic compressibility in [1/Pa].
        """
        raise NotImplementedError(
            "need to implement compressibility() in derived class!")

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.beta_G`.

        Returns
        -------
        shear_modulus : float
            Shear modulus in [Pa].
        """
        raise NotImplementedError(
            "need to implement shear_modulus() in derived class!")

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.v_p`.

        Returns
        -------
        p_wave_velocity : float
            P wave speed in [m/s].
        """
        raise NotImplementedError(
            "need to implement p_wave_velocity() in derived class!")

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.v_phi`.

        Returns
        -------
        bulk sound velocity: float
            Sound velocity in [m/s].
        """
        raise NotImplementedError(
            "need to implement bulk_sound_velocity() in derived class!")

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.v_s`.

        Returns
        -------
        shear_wave_velocity : float
            Wave speed in [m/s].
        """
        raise NotImplementedError(
            "need to implement shear_wave_velocity() in derived class!")

    @material_property
    def grueneisen_parameter(self):
        """
        Returns the grueneisen parameter of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.gr`.

        Returns
        -------
        gr : float
            Grueneisen parameters [unitless].
        """
        raise NotImplementedError(
            "need to implement grueneisen_parameter() in derived class!")

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.alpha`.

        Returns
        -------
        alpha : float
            Thermal expansivity in [1/K].
        """
        raise NotImplementedError(
            "need to implement thermal_expansivity() in derived class!")

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volume of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.C_v`.

        Returns
        -------
        molar_heat_capacity_v : float
            Heat capacity in [J/K/mol].
        """
        raise NotImplementedError(
            "need to implement molar_heat_capacity_v() in derived class!")

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar heat capacity at constant pressure of the mineral.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.C_p`.

        Returns
        -------
        molar_heat_capacity_p : float
            Heat capacity in [J/K/mol].
        """
        raise NotImplementedError(
            "need to implement molar_heat_capacity_p() in derived class!")

    #
    # Aliased properties
    @property
    def P(self):
        """Alias for :func:`~burnman.Material.pressure`"""
        return self.pressure

    @property
    def T(self):
        """Alias for :func:`~burnman.Material.temperature`"""
        return self.temperature

    @property
    def energy(self):
        """Alias for :func:`~burnman.Material.molar_internal_energy`"""
        return self.molar_internal_energy

    @property
    def helmholtz(self):
        """Alias for :func:`~burnman.Material.molar_helmholtz`"""
        return self.molar_helmholtz

    @property
    def gibbs(self):
        """Alias for :func:`~burnman.Material.molar_gibbs`"""
        return self.molar_gibbs

    @property
    def V(self):
        """Alias for :func:`~burnman.Material.molar_volume`"""
        return self.molar_volume

    @property
    def rho(self):
        """Alias for :func:`~burnman.Material.density`"""
        return self.density

    @property
    def S(self):
        """Alias for :func:`~burnman.Material.molar_entropy`"""
        return self.molar_entropy

    @property
    def H(self):
        """Alias for :func:`~burnman.Material.molar_enthalpy`"""
        return self.molar_enthalpy

    @property
    def K_T(self):
        """Alias for :func:`~burnman.Material.isothermal_bulk_modulus`"""
        return self.isothermal_bulk_modulus

    @property
    def K_S(self):
        """Alias for :func:`~burnman.Material.adiabatic_bulk_modulus`"""
        return self.adiabatic_bulk_modulus

    @property
    def beta_T(self):
        """Alias for :func:`~burnman.Material.isothermal_compressibility`"""
        return self.isothermal_compressibility

    @property
    def beta_S(self):
        """Alias for :func:`~burnman.Material.adiabatic_compressibility`"""
        return self.adiabatic_compressibility

    @property
    def G(self):
        """Alias for :func:`~burnman.Material.shear_modulus`"""
        return self.shear_modulus

    @property
    def v_p(self):
        """Alias for :func:`~burnman.Material.p_wave_velocity`"""
        return self.p_wave_velocity

    @property
    def v_phi(self):
        """Alias for :func:`~burnman.Material.bulk_sound_velocity`"""
        return self.bulk_sound_velocity

    @property
    def v_s(self):
        """Alias for :func:`~burnman.Material.shear_wave_velocity`"""
        return self.shear_wave_velocity

    @property
    def gr(self):
        """Alias for :func:`~burnman.Material.grueneisen_parameter`"""
        return self.grueneisen_parameter

    @property
    def alpha(self):
        """Alias for :func:`~burnman.Material.thermal_expansivity`"""
        return self.thermal_expansivity

    @property
    def C_v(self):
        """Alias for :func:`~burnman.Material.molar_heat_capacity_v`"""
        return self.molar_heat_capacity_v

    @property
    def C_p(self):
        """Alias for :func:`~burnman.Material.molar_heat_capacity_p`"""
        return self.molar_heat_capacity_p



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
from .. import constants


def molar_volume_from_unit_cell_volume(unit_cell_v, z):
    """
    Converts a unit cell volume from Angstroms^3 per unitcell,
    to m^3/mol.

    Parameters
    ----------
    unit_cell_v : float
        Unit cell volumes [A^3/unit cell]

    z : float
        Number of formula units per unit cell


    Returns
    -------
    V : float
        Volume [m^3/mol]
    """
    V = unit_cell_v * constants.Avogadro / 1.e30 / z
    return V


def cell_parameters_to_vectors(cell_parameters):
    """
    Converts cell parameters to unit cell vectors.

    Parameters
    ----------
    cell_parameters : 1D numpy array
        An array containing the three lengths of the unit cell vectors [m],
        and the three angles [degrees].
        The first angle (:math:`\\alpha`) corresponds to the angle between the
        second and the third cell vectors, the second (:math:`\\beta`) to the
        angle between the first and third cell vectors, and the third
        (:math:`\\gamma`) to the angle between the first and second vectors.

    Returns
    -------
    M : 2D numpy array
        The three vectors defining the parallelopiped cell [m].
        This function assumes that the first cell vector is colinear with the
        x-axis, and the second is perpendicular to the z-axis, and the third is
        defined in a right-handed sense.
    """
    a, b, c, alpha_deg, beta_deg, gamma_deg = cell_parameters
    alpha = np.radians(alpha_deg)
    beta = np.radians(beta_deg)
    gamma = np.radians(gamma_deg)

    n2 = (np.cos(alpha)-np.cos(gamma)*np.cos(beta))/np.sin(gamma)
    M = np.array([[a, 0, 0],
                  [b*np.cos(gamma), b*np.sin(gamma), 0],
                  [c*np.cos(beta), c*n2, c*np.sqrt(np.sin(beta)**2-n2**2)]])
    return M


def cell_vectors_to_parameters(M):
    """
    Converts unit cell vectors to cell parameters.

    Parameters
    ----------
    M : 2D numpy array
        The three vectors defining the parallelopiped cell [m].
        This function assumes that the first cell vector is colinear with the
        x-axis, the second is perpendicular to the z-axis, and the third is
        defined in a right-handed sense.

    Returns
    -------
    cell_parameters : 1D numpy array
        An array containing the three lengths of the unit cell vectors [m],
        and the three angles [degrees].
        The first angle (:math:`\\alpha`) corresponds to the angle between the
        second and the third cell vectors, the second (:math:`\\beta`) to the
        angle between the first and third cell vectors, and the third
        (:math:`\\gamma`) to the angle between the first and second vectors.
    """

    assert M[0, 1] == 0
    assert M[0, 2] == 0
    assert M[1, 2] == 0

    a = M[0, 0]
    b = np.sqrt(np.power(M[1, 0], 2.) + np.power(M[1, 1], 2.))
    c = (np.sqrt(np.power(M[2, 0], 2.)
                 + np.power(M[2, 1], 2.)
                 + np.power(M[2, 2], 2.)))

    gamma = np.arccos(M[1, 0] / b)
    beta = np.arccos(M[2, 0] / c)
    alpha = np.arccos(M[2, 1]/c*np.sin(gamma) + np.cos(gamma)*np.cos(beta))

    gamma_deg = np.degrees(gamma)
    beta_deg = np.degrees(beta)
    alpha_deg = np.degrees(alpha)

    return np.array([a, b, c, alpha_deg, beta_deg, gamma_deg])



import numpy as np
from sympy.core import S
from sympy.core.numbers import Float, Integer, Rational
from sympy.simplify.simplify import simplify as _simplify

# The following functions are taken from sympy.matrices.utilities
# and sympy.matrices.determinant and sympy.matrices.reductions.
# This bit of sympy appears to be in a high state of flux,
# so we copy the functions here for the time being.


def _iszero(x):
    """Returns True if x is zero."""
    return getattr(x, 'is_zero', None)


def _find_reasonable_pivot(col, iszerofunc=_iszero, simpfunc=_simplify):
    """ Find the lowest index of an item in ``col`` that is
    suitable for a pivot.  If ``col`` consists only of
    Floats, the pivot with the largest norm is returned.
    Otherwise, the first element where ``iszerofunc`` returns
    False is used.  If ``iszerofunc`` doesn't return false,
    items are simplified and retested until a suitable
    pivot is found.

    Returns a 4-tuple
        (pivot_offset, pivot_val, assumed_nonzero, newly_determined)
    where pivot_offset is the index of the pivot, pivot_val is
    the (possibly simplified) value of the pivot, assumed_nonzero
    is True if an assumption that the pivot was non-zero
    was made without being proved, and newly_determined are
    elements that were simplified during the process of pivot
    finding."""

    newly_determined = []
    col = list(col)
    # a column that contains a mix of floats and integers
    # but at least one float is considered a numerical
    # column, and so we do partial pivoting
    if ((all(isinstance(x, (Float, Integer)) for x in col)
         and any(isinstance(x, Float) for x in col))):
        col_abs = [abs(x) for x in col]
        max_value = max(col_abs)
        if iszerofunc(max_value):
            # just because iszerofunc returned True, doesn't
            # mean the value is numerically zero.  Make sure
            # to replace all entries with numerical zeros
            if max_value != 0:
                newly_determined = [(i, 0) for i, x in enumerate(col)
                                    if x != 0]
            return (None, None, False, newly_determined)
        index = col_abs.index(max_value)
        return (index, col[index], False, newly_determined)

    # PASS 1 (iszerofunc directly)
    possible_zeros = []
    for i, x in enumerate(col):
        is_zero = iszerofunc(x)
        # is someone wrote a custom iszerofunc, it may return
        # BooleanFalse or BooleanTrue instead of True or False,
        # so use == for comparison instead of `is`
        if is_zero is False:
            # we found something that is definitely not zero
            return (i, x, False, newly_determined)
        possible_zeros.append(is_zero)

    # by this point, we've found no certain non-zeros
    if all(possible_zeros):
        # if everything is definitely zero, we have
        # no pivot
        return (None, None, False, newly_determined)

    # PASS 2 (iszerofunc after simplify)
    # we haven't found any for-sure non-zeros, so
    # go through the elements iszerofunc couldn't
    # make a determination about and opportunistically
    # simplify to see if we find something
    for i, x in enumerate(col):
        if possible_zeros[i] is not None:
            continue
        simped = simpfunc(x)
        is_zero = iszerofunc(simped)
        if is_zero is True or is_zero is False:
            newly_determined.append((i, simped))
        if is_zero is False:
            return (i, simped, False, newly_determined)
        possible_zeros[i] = is_zero

    # after simplifying, some things that were recognized
    # as zeros might be zeros
    if all(possible_zeros):
        # if everything is definitely zero, we have
        # no pivot
        return (None, None, False, newly_determined)

    # PASS 3 (.equals(0))
    # some expressions fail to simplify to zero, but
    # ``.equals(0)`` evaluates to True.  As a last-ditch
    # attempt, apply ``.equals`` to these expressions
    for i, x in enumerate(col):
        if possible_zeros[i] is not None:
            continue
        if x.equals(S.Zero):
            # ``.iszero`` may return False with
            # an implicit assumption (e.g., ``x.equals(0)``
            # when ``x`` is a symbol), so only treat it
            # as proved when ``.equals(0)`` returns True
            possible_zeros[i] = True
            newly_determined.append((i, S.Zero))

    if all(possible_zeros):
        return (None, None, False, newly_determined)

    # at this point there is nothing that could definitely
    # be a pivot.  To maintain compatibility with existing
    # behavior, we'll assume that an illdetermined thing is
    # non-zero.  We should probably raise a warning in this case
    i = possible_zeros.index(None)
    return (i, col[i], True, newly_determined)


def _row_reduce_list(mat, rows, cols, iszerofunc, simpfunc,
                     normalize_last=True, normalize=True, zero_above=True):
    """Row reduce a flat list representation of a matrix and return a tuple
    (rref_matrix, pivot_cols, swaps) where ``rref_matrix`` is a flat list,
    ``pivot_cols`` are the pivot columns and ``swaps`` are any row swaps that
    were used in the process of row reduction.
    Parameters
    ==========
    mat : list
        list of matrix elements, must be ``rows`` * ``cols`` in length
    rows, cols : integer
        number of rows and columns in flat list representation
    iszerofunc : determines if an entry can be used as a pivot
    simpfunc : used to simplify elements and test if they are
        zero if ``iszerofunc`` returns `None`
    normalize_last : indicates where all row reduction should
        happen in a fraction-free manner and then the rows are
        normalized (so that the pivots are 1), or whether
        rows should be normalized along the way (like the naive
        row reduction algorithm)
    normalize : whether pivot rows should be normalized so that
        the pivot value is 1
    zero_above : whether entries above the pivot should be zeroed.
        If ``zero_above=False``, an echelon matrix will be returned.
    """

    def get_col(i):
        return mat[i::cols]

    def row_swap(i, j):
        mat[i*cols:(i + 1)*cols], mat[j*cols:(j + 1)*cols] = \
            mat[j*cols:(j + 1)*cols], mat[i*cols:(i + 1)*cols]

    def cross_cancel(a, i, b, j):
        """Does the row op row[i] = a*row[i] - b*row[j]"""
        q = (j - i)*cols
        for p in range(i*cols, (i + 1)*cols):
            mat[p] = isimp(a*mat[p] - b*mat[p + q])

    def isimp(x):
        return x
    piv_row, piv_col = 0, 0
    pivot_cols = []
    swaps = []

    # use a fraction free method to zero above and below each pivot
    while piv_col < cols and piv_row < rows:
        pivot_offset, pivot_val, assumed_nonzero, newly_determined = _find_reasonable_pivot(
            get_col(piv_col)[piv_row:], iszerofunc, simpfunc)

        # _find_reasonable_pivot may have simplified some things
        # in the process.  Let's not let them go to waste
        for (offset, val) in newly_determined:
            offset += piv_row
            mat[offset*cols + piv_col] = val

        if pivot_offset is None:
            piv_col += 1
            continue

        pivot_cols.append(piv_col)
        if pivot_offset != 0:
            row_swap(piv_row, pivot_offset + piv_row)
            swaps.append((piv_row, pivot_offset + piv_row))

        # if we aren't normalizing last, we normalize
        # before we zero the other rows
        if normalize_last is False:
            i, j = piv_row, piv_col
            mat[i*cols + j] = S.One
            for p in range(i*cols + j + 1, (i + 1)*cols):
                mat[p] = isimp(mat[p] / pivot_val)
            # after normalizing, the pivot value is 1
            pivot_val = S.One

        # zero above and below the pivot
        for row in range(rows):
            # don't zero our current row
            if row == piv_row:
                continue
            # don't zero above the pivot unless we're told.
            if zero_above is False and row < piv_row:
                continue
            # if we're already a zero, don't do anything
            val = mat[row*cols + piv_col]
            if iszerofunc(val):
                continue

            cross_cancel(pivot_val, row, val, piv_row)
        piv_row += 1

    # normalize each row
    if normalize_last is True and normalize is True:
        for piv_i, piv_j in enumerate(pivot_cols):
            pivot_val = mat[piv_i*cols + piv_j]
            mat[piv_i*cols + piv_j] = S.One
            for p in range(piv_i*cols + piv_j + 1, (piv_i + 1)*cols):
                mat[p] = isimp(mat[p] / pivot_val)

    return mat, tuple(pivot_cols), tuple(swaps)


# This functions is a candidate for caching
# if it gets implemented for matrices.
def row_reduce(M, iszerofunc=lambda x: x.is_zero,
               simpfunc=lambda x: Rational(x).limit_denominator(1000),
               normalize_last=True,
               normalize=True, zero_above=True):

    mat, pivot_cols, swaps = _row_reduce_list(list(M), M.rows, M.cols,
                                              iszerofunc, simpfunc,
                                              normalize_last=normalize_last,
                                              normalize=normalize,
                                              zero_above=zero_above)

    return M._new(M.rows, M.cols, mat), pivot_cols, swaps


def independent_row_indices(m, iszerofunc=lambda x: x.is_zero,
                            simpfunc=lambda x: Rational(x).limit_denominator(1000)):

    _, pivots, swaps = row_reduce(m, iszerofunc, simpfunc)
    indices = np.array(range(len(m)))
    for swap in np.array(swaps):
        indices[swap] = indices[swap[::-1]]
    return indices[:len(pivots)]



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
HHPH_2013
^^^^^^^^^

Minerals from Holland et al. (2013) and references therein.
The values in this document are all in S.I. units,
unlike those in the original paper.
File autogenerated using HHPHdata_to_burnman.py.
"""
from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..classes.solidsolution import SolidSolution
from ..classes.solutionmodel import *
from ..tools.chemistry import dictionarize_formula, formula_mass


"""
ENDMEMBERS
"""


class fo (Mineral):

    def __init__(self):
        formula = 'Mg2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'fo',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -2172450.0,
            'S_0': 95.1,
            'V_0': 4.366e-05,
            'Cp': [233.3, 0.001494, -603800.0, -1869.7],
            'a_0': 2.85e-05,
            'K_0': 1.285e+11,
            'Kprime_0': 3.84,
            'Kdprime_0': -3e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 530.0}
        Mineral.__init__(self)


class fa (Mineral):

    def __init__(self):
        formula = 'Fe2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'fa',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1477740.0,
            'S_0': 151.0,
            'V_0': 4.631e-05,
            'Cp': [201.1, 0.01733, -1960600.0, -900.9],
            'a_0': 2.82e-05,
            'K_0': 1.256e+11,
            'Kprime_0': 4.68,
            'Kdprime_0': -3.7e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 640.0}
        Mineral.__init__(self)


class mwd (Mineral):

    def __init__(self):
        formula = 'Mg2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'mwd',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -2138080.0,
            'S_0': 93.9,
            'V_0': 4.051e-05,
            'Cp': [208.7, 0.003942, -1709500.0, -1302.8],
            'a_0': 2.37e-05,
            'K_0': 1.726e+11,
            'Kprime_0': 3.84,
            'Kdprime_0': -2.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 620.0}
        Mineral.__init__(self)


class fwd (Mineral):

    def __init__(self):
        formula = 'Fe2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'fwd',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1467920.0,
            'S_0': 146.0,
            'V_0': 4.321e-05,
            'Cp': [201.1, 0.01733, -1960600.0, -900.9],
            'a_0': 2.73e-05,
            'K_0': 1.69e+11,
            'Kprime_0': 4.35,
            'Kdprime_0': -2.6e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 900.0}
        Mineral.__init__(self)


class mrw (Mineral):

    def __init__(self):
        formula = 'Mg2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'mrw',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -2126840.0,
            'S_0': 90.0,
            'V_0': 3.949e-05,
            'Cp': [213.3, 0.00269, -1410400.0, -1495.9],
            'a_0': 2.01e-05,
            'K_0': 1.781e+11,
            'Kprime_0': 4.35,
            'Kdprime_0': -2.4e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 630.0}
        Mineral.__init__(self)


class frw (Mineral):

    def __init__(self):
        formula = 'Fe2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'frw',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1471760.0,
            'S_0': 140.0,
            'V_0': 4.203e-05,
            'Cp': [166.8, 0.04261, -1705400.0, -541.4],
            'a_0': 2.22e-05,
            'K_0': 1.977e+11,
            'Kprime_0': 4.92,
            'Kdprime_0': -2.5e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 710.0}
        Mineral.__init__(self)


class mpv (Mineral):

    def __init__(self):
        formula = 'MgSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'mpv',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1442310.0,
            'S_0': 62.6,
            'V_0': 2.445e-05,
            'Cp': [149.3, 0.002918, -2983000.0, -799.1],
            'a_0': 1.87e-05,
            'K_0': 2.51e+11,
            'Kprime_0': 4.14,
            'Kdprime_0': -1.6e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 470.0}
        Mineral.__init__(self)


class fpv (Mineral):

    def __init__(self):
        formula = 'FeSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'fpv',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1082910.0,
            'S_0': 95.0,
            'V_0': 2.534e-05,
            'Cp': [133.2, 0.01083, -3661400.0, -314.7],
            'a_0': 1.87e-05,
            'K_0': 2.81e+11,
            'Kprime_0': 4.14,
            'Kdprime_0': -1.6e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 760.0}
        Mineral.__init__(self)


class apv (Mineral):

    def __init__(self):
        formula = 'AlAlO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'apv',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1619990.0,
            'S_0': 51.8,
            'V_0': 2.54e-05,
            'Cp': [139.5, 0.00589, -2460600.0, -589.2],
            'a_0': 1.8e-05,
            'K_0': 2.03e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 770.0}
        Mineral.__init__(self)


class npv (Mineral):

    def __init__(self):
        formula = 'Na0.5Al0.5SiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'npv',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1365000.0,
            'S_0': 63.0,
            'V_0': 2.334e-05,
            'Cp': [135.0, 0.00846, -1850300.0, -600.8],
            'a_0': 1.8e-05,
            'K_0': 2.03e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 10240.0}
        Mineral.__init__(self)


class cpv (Mineral):

    def __init__(self):
        formula = 'CaSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'cpv',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1533590.0,
            'S_0': 74.5,
            'V_0': 2.745e-05,
            'Cp': [159.3, 0.0, -967300.0, -1075.4],
            'a_0': 2e-05,
            'K_0': 2.36e+11,
            'Kprime_0': 3.9,
            'Kdprime_0': -1.6e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 1090.0}
        Mineral.__init__(self)


class mak (Mineral):

    def __init__(self):
        formula = 'MgSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'mak',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1489610.0,
            'S_0': 59.3,
            'V_0': 2.635e-05,
            'Cp': [147.8, 0.002015, -2395000.0, -801.8],
            'a_0': 2.12e-05,
            'K_0': 2.11e+11,
            'Kprime_0': 4.55,
            'Kdprime_0': -2.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 420.0}
        Mineral.__init__(self)


class fak (Mineral):

    def __init__(self):
        formula = 'FeSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'fak',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1142130.0,
            'S_0': 91.5,
            'V_0': 2.76e-05,
            'Cp': [100.3, 0.013328, -4364900.0, 419.8],
            'a_0': 2.12e-05,
            'K_0': 2.18e+11,
            'Kprime_0': 4.55,
            'Kdprime_0': -2.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 9630.0}
        Mineral.__init__(self)


class maj (Mineral):

    def __init__(self):
        formula = 'Mg4Si4O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'maj',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -6041550.0,
            'S_0': 260.2,
            'V_0': 0.00011457,
            'Cp': [713.6, -0.000997, -1158200.0, -6622.3],
            'a_0': 1.83e-05,
            'K_0': 1.6e+11,
            'Kprime_0': 4.56,
            'Kdprime_0': -2.8e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 2260.0}
        Mineral.__init__(self)


class nagt (Mineral):

    def __init__(self):
        formula = 'Mg2NaAlSiSi3O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'nagt',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -5985000.0,
            'S_0': 260.6,
            'V_0': 0.0001109,
            'Cp': [620.8, 0.0112, -3755900.0, -4421.3],
            'a_0': 2.1e-05,
            'K_0': 1.7e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -2.3e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 5120.0}
        Mineral.__init__(self)


class py (Mineral):

    def __init__(self):
        formula = 'Mg3Al2Si3O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'py',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -6281770.0,
            'S_0': 269.5,
            'V_0': 0.00011313,
            'Cp': [633.5, 0.0, -5196100.0, -4315.2],
            'a_0': 2.37e-05,
            'K_0': 1.743e+11,
            'Kprime_0': 4.05,
            'Kdprime_0': -2.3e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 990.0}
        Mineral.__init__(self)


class alm (Mineral):

    def __init__(self):
        formula = 'Fe3Al2Si3O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'alm',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -5260750.0,
            'S_0': 342.0,
            'V_0': 0.00011525,
            'Cp': [677.3, 0.0, -3772700.0, -5044.0],
            'a_0': 2.12e-05,
            'K_0': 1.9e+11,
            'Kprime_0': 2.98,
            'Kdprime_0': -1.6e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 1200.0}
        Mineral.__init__(self)


class gr (Mineral):

    def __init__(self):
        formula = 'Ca3Al2Si3O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'gr',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -6643050.0,
            'S_0': 255.0,
            'V_0': 0.00012535,
            'Cp': [626.0, 0.0, -5779200.0, -4002.9],
            'a_0': 2.2e-05,
            'K_0': 1.72e+11,
            'Kprime_0': 5.53,
            'Kdprime_0': -3.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 1370.0}
        Mineral.__init__(self)


class en (Mineral):

    def __init__(self):
        formula = 'Mg2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'en',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -3090100.0,
            'S_0': 132.5,
            'V_0': 6.262e-05,
            'Cp': [356.2, -0.00299, -596900.0, -3185.3],
            'a_0': 2.27e-05,
            'K_0': 1.059e+11,
            'Kprime_0': 8.65,
            'Kdprime_0': -8.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 620.0}
        Mineral.__init__(self)


class cen (Mineral):

    def __init__(self):
        formula = 'Mg2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'cen',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -3090990.0,
            'S_0': 132.0,
            'V_0': 6.264e-05,
            'Cp': [306.0, -0.003793, -3041700.0, -1852.1],
            'a_0': 2.11e-05,
            'K_0': 1.059e+11,
            'Kprime_0': 8.65,
            'Kdprime_0': -8.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 620.0}
        Mineral.__init__(self)


class hen (Mineral):

    def __init__(self):
        formula = 'Mg2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'hen',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -3082610.0,
            'S_0': 131.7,
            'V_0': 6.099e-05,
            'Cp': [356.2, -0.00299, -596900.0, -3185.3],
            'a_0': 2.26e-05,
            'K_0': 1.5e+11,
            'Kprime_0': 5.5,
            'Kdprime_0': -3.6e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 620.0}
        Mineral.__init__(self)


class hfs (Mineral):

    def __init__(self):
        formula = 'Fe2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'hfs',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -2380810.0,
            'S_0': 189.0,
            'V_0': 6.405e-05,
            'Cp': [398.7, -0.006579, 1290100.0, -4058.0],
            'a_0': 2.37e-05,
            'K_0': 1.5e+11,
            'Kprime_0': 5.5,
            'Kdprime_0': -3.6e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 790.0}
        Mineral.__init__(self)


class fs (Mineral):

    def __init__(self):
        formula = 'Fe2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'fs',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -2388760.0,
            'S_0': 189.9,
            'V_0': 6.592e-05,
            'Cp': [398.7, -0.006579, 1290100.0, -4058.0],
            'a_0': 3.26e-05,
            'K_0': 1.01e+11,
            'Kprime_0': 4.08,
            'Kdprime_0': -4e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 750.0}
        Mineral.__init__(self)


class mgts (Mineral):

    def __init__(self):
        formula = 'MgAl2SiO6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'mgts',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -3196600.0,
            'S_0': 131.0,
            'V_0': 6.05e-05,
            'Cp': [371.4, -0.004082, -398400.0, -3547.1],
            'a_0': 2.17e-05,
            'K_0': 1.028e+11,
            'Kprime_0': 8.55,
            'Kdprime_0': -8.3e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 690.0}
        Mineral.__init__(self)


class di (Mineral):

    def __init__(self):
        formula = 'CaMgSi2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'di',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -3201820.0,
            'S_0': 142.9,
            'V_0': 6.619e-05,
            'Cp': [314.5, 4.1e-05, -2745900.0, -2020.1],
            'a_0': 2.73e-05,
            'K_0': 1.192e+11,
            'Kprime_0': 5.19,
            'Kdprime_0': -4.4e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 580.0}
        Mineral.__init__(self)


class hed (Mineral):

    def __init__(self):
        formula = 'CaFeSi2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'hed',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -2842120.0,
            'S_0': 175.0,
            'V_0': 6.795e-05,
            'Cp': [340.2, 0.000812, -1047800.0, -2646.7],
            'a_0': 2.38e-05,
            'K_0': 1.192e+11,
            'Kprime_0': 3.97,
            'Kdprime_0': -3.3e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 880.0}
        Mineral.__init__(self)


class jd (Mineral):

    def __init__(self):
        formula = 'NaAlSiO6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'jd',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -3025290.0,
            'S_0': 133.5,
            'V_0': 6.04e-05,
            'Cp': [319.4, 0.003616, -1173900.0, -2469.5],
            'a_0': 2.1e-05,
            'K_0': 1.281e+11,
            'Kprime_0': 3.81,
            'Kdprime_0': -3e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 1520.0}
        Mineral.__init__(self)


class cats (Mineral):

    def __init__(self):
        formula = 'CaAl2SiO6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'cats',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -3310110.0,
            'S_0': 135.0,
            'V_0': 6.356e-05,
            'Cp': [347.6, -0.006974, -1781600.0, -2757.5],
            'a_0': 2.08e-05,
            'K_0': 1.192e+11,
            'Kprime_0': 5.19,
            'Kdprime_0': -4.4e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 750.0}
        Mineral.__init__(self)


class stv (Mineral):

    def __init__(self):
        formula = 'SiO2'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'stv',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -876820.0,
            'S_0': 24.0,
            'V_0': 1.401e-05,
            'Cp': [68.1, 0.00601, -1978200.0, -82.1],
            'a_0': 1.58e-05,
            'K_0': 3.09e+11,
            'Kprime_0': 4.6,
            'Kdprime_0': -1.5e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 420.0}
        Mineral.__init__(self)


class macf (Mineral):

    def __init__(self):
        formula = 'MgAl2O4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'macf',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -2246420.0,
            'S_0': 80.0,
            'V_0': 3.614e-05,
            'Cp': [200.0, 0.006252, -2996400.0, -888.4],
            'a_0': 1.93e-05,
            'K_0': 2.12e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -1.7e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 1080.0}
        Mineral.__init__(self)


class mscf (Mineral):

    def __init__(self):
        formula = 'Mg2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'mscf',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -2061130.0,
            'S_0': 87.5,
            'V_0': 3.649e-05,
            'Cp': [213.3, 0.00269, -1410400.0, -1495.9],
            'a_0': 2.01e-05,
            'K_0': 1.85e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -1.7e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 1340.0}
        Mineral.__init__(self)


class fscf (Mineral):

    def __init__(self):
        formula = 'Fe2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'fscf',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1405500.0,
            'S_0': 143.4,
            'V_0': 3.914e-05,
            'Cp': [181.1, 0.018526, -2767200.0, -527.1],
            'a_0': 2.01e-05,
            'K_0': 1.85e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -1.7e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 10240.0}
        Mineral.__init__(self)


class nacf (Mineral):

    def __init__(self):
        formula = 'NaAlSiO6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'nacf',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1965550.0,
            'S_0': 110.0,
            'V_0': 3.631e-05,
            'Cp': [272.7, -0.012398, 0.0, -2763.1],
            'a_0': 2.1e-05,
            'K_0': 1.85e+11,
            'Kprime_0': 4.6,
            'Kdprime_0': -2.5e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 3440.0}
        Mineral.__init__(self)


class cacf (Mineral):

    def __init__(self):
        formula = 'CaAl2O4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'cacf',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -2325600.0,
            'S_0': 87.6,
            'V_0': 3.976e-05,
            'Cp': [191.9, 0.009563, -3211300.0, -640.2],
            'a_0': 1.93e-05,
            'K_0': 1.9e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -2.1e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 10240.0}
        Mineral.__init__(self)


class manal (Mineral):

    def __init__(self):
        formula = 'Mg3Al6O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'manal',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -6796630.0,
            'S_0': 250.0,
            'V_0': 0.00011166,
            'Cp': [600.0, 0.018756, -8989200.0, -2665.2],
            'a_0': 1.93e-05,
            'K_0': 1.84e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -2.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 5120.0}
        Mineral.__init__(self)


class nanal (Mineral):

    def __init__(self):
        formula = 'NaMg2SiAl5O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'nanal',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -6610270.0,
            'S_0': 280.0,
            'V_0': 0.00011322,
            'Cp': [672.7, 0.000106, -5992800.0, -4539.9],
            'a_0': 2.01e-05,
            'K_0': 1.84e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -2.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 5120.0}
        Mineral.__init__(self)


class msnal (Mineral):

    def __init__(self):
        formula = 'MgMg2Si3Mg3O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'msnal',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -6172380.0,
            'S_0': 272.5,
            'V_0': 0.00011061,
            'Cp': [639.9, 0.00807, -4231200.0, -4487.7],
            'a_0': 2.1e-05,
            'K_0': 1.85e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -2.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 5120.0}
        Mineral.__init__(self)


class fsnal (Mineral):

    def __init__(self):
        formula = 'FeFe2Si3Fe3O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'fsnal',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -4146000.0,
            'S_0': 440.2,
            'V_0': 0.00011856,
            'Cp': [543.3, 0.055578, -8301600.0, -1581.3],
            'a_0': 2.1e-05,
            'K_0': 1.85e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -2.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 10240.0}
        Mineral.__init__(self)


class canal (Mineral):

    def __init__(self):
        formula = 'CaMg2Al6O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'canal',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -6840000.0,
            'S_0': 257.6,
            'V_0': 0.00011159,
            'Cp': [591.9, 0.022067, -9204100.0, -2417.0],
            'a_0': 1.93e-05,
            'K_0': 1.77e+11,
            'Kprime_0': 4.0,
            'Kdprime_0': -2.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 5120.0}
        Mineral.__init__(self)


class per (Mineral):

    def __init__(self):
        formula = 'MgO'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'per',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -601570.0,
            'S_0': 26.5,
            'V_0': 1.125e-05,
            'Cp': [60.5, 0.000362, -535800.0, -299.2],
            'a_0': 3.11e-05,
            'K_0': 1.616e+11,
            'Kprime_0': 3.95,
            'Kdprime_0': -2.4e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 260.0}
        Mineral.__init__(self)


class fper (Mineral):

    def __init__(self):
        formula = 'FeO'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'fper',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -262240.0,
            'S_0': 58.6,
            'V_0': 1.206e-05,
            'Cp': [44.4, 0.00828, -1214200.0, 185.2],
            'a_0': 3.22e-05,
            'K_0': 1.52e+11,
            'Kprime_0': 4.9,
            'Kdprime_0': -3.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 680.0}
        Mineral.__init__(self)


class cor (Mineral):

    def __init__(self):
        formula = 'Al2O3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'cor',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1675250.0,
            'S_0': 50.9,
            'V_0': 2.558e-05,
            'Cp': [139.5, 0.00589, -2460600.0, -589.2],
            'a_0': 1.8e-05,
            'K_0': 2.54e+11,
            'Kprime_0': 4.34,
            'Kdprime_0': -1.7e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 700.0}
        Mineral.__init__(self)


class mcor (Mineral):

    def __init__(self):
        formula = 'MgSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'mcor',
            'formula': formula,
            'equation_of_state': 'hp_tmt',
            'H_0': -1468000.0,
            'S_0': 59.3,
            'V_0': 2.635e-05,
            'Cp': [147.8, 0.002015, -2395000.0, -801.8],
            'a_0': 2.12e-05,
            'K_0': 2.11e+11,
            'Kprime_0': 4.55,
            'Kdprime_0': -2.2e-11,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_H_0': 880.0}
        Mineral.__init__(self)



from __future__ import print_function
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
from scipy.integrate import odeint
from scipy.integrate import quad
from scipy.interpolate import UnivariateSpline, interp1d
from burnman import constants
from burnman.tools import geotherm
import warnings

from .material import Material, material_property


class Layer(object):
    """
    The base class for a planetary layer.
    The user needs to set the following before properties can be computed:

    - set_material(), which sets the material of the layer,
      e.g. a mineral, solid_solution, or composite
    - set_temperature_mode(), either predefine, or set to an adiabatic profile
    - set_pressure_mode(), to set the self-consistent pressure
      (with user-defined option the pressures can be overwritten).
      To set the self-consistent pressure the pressure at the top and the
      gravity at the bottom of the layer need to be set.
    - make(), computes the self-consistent part of the layer and starts the
      settings to compute properties within the layer

    Note that the entire planet this layer sits in is not necessarily
    self-consistent, as the pressure at the top of the layer is a
    function of the density within the layer (through the gravity).
    Entire planets can be computed self-consistently with the planet class.
    Properties will be returned at the pre-defined radius array,
    although the evaluate() function can take a newly defined depthlist
    and values are interpolated between these (sufficient sampling of the layer
    is needed for this to be accurate).
    """

    def __init__(self, name=None, radii=None, verbose=False):
        self.name = name
        assert np.all(np.diff(radii) > 0)
        self.radii = radii
        self.outer_radius = max(self.radii)
        self.inner_radius = min(self.radii)
        self.thickness = self.outer_radius - self.inner_radius
        self.n_slices = len(self.radii)
        self.verbose = verbose
        self._cached = {}
        self._pressures = None
        self._temperatures = None
        self.sublayers = None
        self.material = None
        self.pressure_mode = 'self-consistent'
        self.temperature_mode = None

    def __str__(self):
        """
        Prints details of the layer
        """
        writing = (f'The {self.name} is made of {self.material.name}'
                   f' with {self.temperature_mode} temperatures and '
                   f'{self.pressure_mode} pressures\n')
        return writing

    def reset(self):
        """
        Resets all cached material properties.
        It is typically not required for the user to call this function.
        """
        self._cached = {}
        self._pressures = None
        self._temperatures = None
        self.sublayers = None

    def set_material(self, material):
        """
        Set the material of a Layer with a Material
        """
        assert(isinstance(material, Material))
        self.material = material
        self.reset()

    def set_temperature_mode(self, temperature_mode='adiabatic',
                             temperatures=None, temperature_top=None):
        """
        Sets temperatures within the layer as user-defined values or as
        a (potentially perturbed) adiabat.

        Parameters
        ----------
        temperature_mode : string
            This can be set to 'user-defined', 'adiabatic',
            or 'perturbed-adiabatic'. 'user-defined' fixes the temperature
            with the profile input by the user. 'adiabatic' self-consistently
            computes the adiabat when setting the state of the layer.
            'perturbed-adiabatic' adds the user input array to the adiabat.
            This allows the user to apply boundary layers (for example).
        temperatures : array of float
            The desired fixed temperatures in [K].
            Should have same length as defined radii in layer.
        temperature_top : float
            Temperature at the top for an adiabat.
        """
        self.reset()
        assert(temperature_mode == 'user-defined'
               or temperature_mode == 'adiabatic'
               or temperature_mode == 'perturbed-adiabatic')

        self.temperature_mode = temperature_mode

        if ((temperature_mode == 'user-defined'
             or temperature_mode == 'perturbed-adiabatic')):
            assert(len(temperatures) == len(self.radii))
            self.usertemperatures = temperatures
        else:
            self.usertemperatures = np.zeros_like(self.radii)

        if ((temperature_mode == 'adiabatic'
             or temperature_mode == 'perturbed-adiabatic')):
            self.temperature_top = temperature_top
        else:
            self.temperature_top = None

    def set_pressure_mode(self, pressure_mode='self-consistent',
                          pressures=None, gravity_bottom=None,
                          pressure_top=None, n_max_iterations=50,
                          max_delta=1.e-5):
        """
        Sets the pressure mode of the layer,
        which can either be 'user-defined', or 'self-consistent'.

        Parameters
        ----------
        pressure_mode : string
            This can be set to 'user-defined' or 'self-consistent'.
            'user-defined' fixes the pressures with the profile input
            by the user in the 'pressures' argument.
            'self-consistent' forces Layer to calculate pressures
            self-consistently. If this is selected, the user will need
            to supply values for the gravity_bottom [m/s^2]
            and pressure_top [Pa] arguments.
        pressures : array of floats
            Pressures [Pa] to set layer to
            (if the 'user-defined' pressure_mode has been selected).
            The array should be the same length as
            the layers user-defined radii array.
        pressure_top : float
            Pressure [Pa] at the top of the layer.
        gravity_bottom : float
            gravity [m/s^2] at the bottom of the layer.
        n_max_iterations : integer
            Maximum number of iterations to reach
            self-consistent pressures (default = 50)
        max_delta : float
            Relative update to the highest pressure in the layer between
            iterations to stop iterations (default = 1.e-5)
        """
        self.reset()
        assert(pressure_mode == 'user-defined'
               or pressure_mode == 'self-consistent')
        self.pressure_mode = pressure_mode

        assert(gravity_bottom is not None)
        self.gravity_bottom = gravity_bottom

        if pressure_mode == 'user-defined':
            assert(pressures is not None)
            assert(len(pressures) == len(self.radii))
            self.pressures = pressures
            warnings.warn("By setting the pressures in Layer they "
                          "are unlikely to be self-consistent")
        elif pressure_mode == 'self-consistent':
            self.pressure_top = pressure_top
            self.n_max_iterations = n_max_iterations
            self.max_delta = max_delta
        else:
            raise NotImplementedError(f'pressure mode {pressure_mode} '
                                      'not recognised')

    def make(self):
        """
        This routine needs to be called before evaluating any properties.
        If pressures and temperatures are not user-defined, they
        are computed here. This method also initializes an array of copied
        materials from which properties can be computed.
        """
        self.reset()
        if not hasattr(self, 'material'):
            raise AttributeError('You must set_material() for the layer '
                                 'before running make().')
        if not hasattr(self, 'temperature_mode'):
            raise AttributeError('You must set_temperature_mode() for the '
                                 'layer before running make().')
        if not hasattr(self, 'pressure_mode'):
            raise AttributeError('You must set_pressure_mode() for the layer '
                                 'before running make().')

        if self.pressure_mode == 'user-defined':
            self.temperatures = self._evaluate_temperature(
                self.pressures, self.temperature_top)
        elif self.pressure_mode == 'self-consistent':
            new_press = (self.pressure_top
                         + (-self.radii + max(self.radii))
                         * 1.e3)  # initial pressure curve guess
            temperatures = self._evaluate_temperature(
                new_press, self.temperature_top)
            # Make it self-consistent!!!
            i = 0

            while i < self.n_max_iterations:
                i += 1
                ref_press = new_press
                new_grav, new_press = self._evaluate_eos(
                    new_press, temperatures, self.gravity_bottom,
                    self.pressure_top)
                temperatures = self._evaluate_temperature(
                    new_press, self.temperature_top)
                rel_err = abs(
                    (max(ref_press) - max(new_press)) / max(new_press))
                if self.verbose:
                    print(f'Iteration {i:0d} maximum relative pressure error: '
                          f'{rel_err:.1f}')

                if rel_err < self.max_delta:
                    break

            self.pressures = new_press
            self.temperatures = temperatures
        else:
            raise NotImplementedError('pressure mode not recognised')

        self.sublayers = []
        for r in range(len(self.radii)):
            self.sublayers.append(self.material.copy())
            self.sublayers[r].set_state(self.pressures[r],
                                        self.temperatures[r])

    def evaluate(self, properties, radlist=None, radius_planet=None):
        """
        Function that is used to evaluate properties
        across the layer. If radlist is not defined, values are
        returned at the internal radlist.
        If asking for different radii than the internal radlist,
        pressure and temperature values are interpolated and the
        layer material evaluated at those pressures and
        temperatures.

        Parameters
        ----------
        properties : list of strings
            List of properties to evaluate
        radlist : array of floats
            Radii to evaluate properties at. If left empty,
            internal radii list is used.
        planet_radius : float
            Planet outer radius. Used only to calculate depth.

        Returns
        -------
        property_array : numpy array
            1D or 2D array of requested properties
            (1D if only one property was requested)
        """

        if radlist is None:
            values = np.empty([len(properties), len(self.radii)])
            for i, prop in enumerate(properties):
                if prop == 'depth':
                    values[i] = radius_planet - self.radii
                else:
                    try:
                        values[i] = getattr(self, prop)
                    except:
                        values[i] = np.array([getattr(self.sublayers[i], prop)
                                              for i
                                              in range(len(self.sublayers))])
        else:
            func_p = interp1d(self.radii, self.pressures)
            pressures = func_p(radlist)
            func_t = interp1d(self.radii, self.temperatures)
            temperatures = func_t(radlist)
            values = np.empty([len(properties), len(radlist)])
            for i, prop in enumerate(properties):
                if prop == 'depth':
                    values[i] = radius_planet - radlist
                else:
                    try:
                        values[i] = self.material.evaluate([prop],
                                                           pressures,
                                                           temperatures)
                    except:
                        func_prop = interp1d(self.radii, getattr(self, prop))
                        values[i] = func_prop(radlist)

        if values.shape[0] == 1:
            values = values[0]
        return values

    def _evaluate_temperature(self, pressures=None, temperature_top=None):
        """
        Returns the temperatures of the layer for given pressures.
        Used by make()
        """
        if ((self.temperature_mode == 'adiabatic'
             or self.temperature_mode == 'perturbed-adiabatic')):
            adiabat = geotherm.adiabatic(pressures[::-1], temperature_top,
                                         self.material)[::-1]
        else:
            adiabat = np.zeros_like(self.radii)
        return adiabat + self.usertemperatures

    def _evaluate_eos(self, pressures, temperatures,
                      gravity_bottom, pressure_top):
        """
        Returns updated gravity and pressure
        make() loops over this until consistency is achieved.
        """
        [density] = self.material.evaluate(
            ['density'], pressures, temperatures)
        grav = self._compute_gravity(density, gravity_bottom)
        press = self._compute_pressure(density, grav, pressure_top)
        return grav, press

    # Functions needed to compute self-consistent radii-pressures
    def _compute_gravity(self, density, gravity_bottom):
        """
        Computes the gravity of a layer
        Used by _evaluate_eos()
        """
        # Create a spline fit of density as a function of radius
        rhofunc = UnivariateSpline(self.radii, density)
        # Numerically integrate Poisson's equation

        def poisson(p, x): return 4.0 * np.pi * \
            constants.G * rhofunc(x) * x * x
        grav = np.ravel(odeint(poisson,
                               gravity_bottom * self.radii[0] * self.radii[0],
                               self.radii))

        if self.radii[0] == 0:
            grav[0] = 0
            grav[1:] = grav[1:] / self.radii[1:] / self.radii[1:]
        else:
            grav[:] = grav[:] / self.radii[:] / self.radii[:]
        return grav

    def _compute_pressure(self, density, gravity, pressure_top):
        """
        Calculate the pressure profile based on density and gravity.
        This integrates the equation for hydrostatic equilibrium P = rho g z.
        Used by _evaluate_eos()
        """
        # flip radius, density and gravity to increasing pressure
        depthfromtop = -self.radii[::-1] + max(self.radii)
        density = density[::-1]
        gravity = gravity[::-1]
        # Make a spline fit of density as a function of depth
        rhofunc = UnivariateSpline(depthfromtop, density)
        # Make a spline fit of gravity as a function of depth
        gfunc = UnivariateSpline(depthfromtop, gravity)

        # integrate the hydrostatic equation
        pressure = np.ravel(odeint((lambda p, x: gfunc(x) * rhofunc(x)),
                                   pressure_top, depthfromtop))

        return pressure[::-1]

    @property
    def mass(self):
        """
        Calculates the mass of the layer [kg]
        """
        mass = 0.0
        radii = self.radii
        density = self.evaluate(['density'])
        rhofunc = UnivariateSpline(radii, density)
        mass = np.abs(quad(lambda r: 4 * np.pi * rhofunc(r)
                           * r * r, radii[0], radii[-1])[0])
        return mass

    @property
    def moment_of_inertia(self):
        """
        Returns the moment of inertia of the layer [kg m^2]
        """
        moment = 0.0
        radii = self.radii
        density = self.evaluate(['density'])
        rhofunc = UnivariateSpline(radii, density)
        moment = np.abs(quad(lambda r: 8.0 / 3.0 * np.pi * rhofunc(r)
                             * r * r * r * r, radii[0], radii[-1])[0])
        return moment

    @property
    def gravity(self):
        """
        Returns gravity profile of the layer [m s^(-2)]
        """
        return self._compute_gravity(self.density, self.gravity_bottom)

    @property
    def bullen(self):
        """
        Returns the Bullen parameter across the layer.
        The Bullen parameter assess if compression as a function of pressure is
        like homogeneous, adiabatic compression.
        Bullen parameter =1  , homogeneous, adiabatic compression
        Bullen parameter > 1 , more compressed with pressure,
        e.g. across phase transitions Bullen parameter < 1,
        less compressed with pressure, e.g. across a boundary layer.
        """
        kappa = (self.bulk_sound_velocity
                 * self.bulk_sound_velocity * self.density)
        phi = self.bulk_sound_velocity * self.bulk_sound_velocity
        try:
            dkappadP = np.gradient(kappa, edge_order=2) / \
                       np.gradient(self.pressures, edge_order=2)
            dphidr = (np.gradient(phi, edge_order=2)
                      / np.gradient(self.radii, edge_order=2)
                      / self.gravity)
        except:
            dkappadP = np.gradient(kappa) / np.gradient(self.pressures)
            dphidr = np.gradient(phi) / np.gradient(self.radii) / self.gravity
        bullen = dkappadP + dphidr
        return bullen

    @property
    def brunt_vasala(self):
        """
        Returns the brunt-vasala (or buoyancy) frequency, N, across the layer.
        This frequency assess the stabilty of the layer:
        N < 0, fluid will convect
        N= 0, fluid is neutral
        N > 0, fluid is stabily stratified.
        """
        kappa = (self.bulk_sound_velocity
                 * self.bulk_sound_velocity * self.density)
        brunt_vasala = self.density * self.gravity * \
            self.gravity * (self.bullen - 1.) / kappa
        return brunt_vasala

    @property
    def pressure(self):
        """
        Returns current pressures across the layer that was set
        with :func:`~burnman.Material.set_state`.

        Aliased with :func:`~burnman.Material.P`.

        Returns
        -------
        pressure : array of floats
            Pressures in [Pa] at the predefined radii.
        """
        return self.pressures

    @property
    def temperature(self):
        """
        Returns current temperature  across the layer that was set with
        :func:`~burnman.Material.set_state`.

        - Aliased with :func:`~burnman.Material.T`.

        Returns
        -------
        temperature : array of floats
            Temperatures in [K] at the predefined radii.
        """
        return self.temperatures

    @material_property
    def molar_internal_energy(self):
        """
        Returns the molar internal energies across the layer.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.energy`.

        Returns
        -------
        molar_internal_energy : array of floats
            The internal energies in [J/mol] at the predefined radii.
        """
        return np.array([self.sublayers[i].molar_internal_energy
                         for i in range(len(self.sublayers))])

    @material_property
    def molar_gibbs(self):
        """
        Returns the molar Gibbs free energies across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.gibbs`.

        Returns
        -------
        molar_gibbs : array of floats
            Gibbs free energies in [J/mol] at the predefined radii.
        """
        return np.array([self.sublayers[i].molar_gibbs
                         for i in range(len(self.sublayers))])

    @material_property
    def molar_helmholtz(self):
        """
        Returns the molar Helmholtz free energies across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.helmholtz`.

        Returns
        -------
        molar_helmholtz : array of floats
            Helmholtz free energies in [J/mol] at the predefined radii.
        """
        return np.array([self.sublayers[i].molar_helmholtz
                         for i in range(len(self.sublayers))])

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the layer.

        Needs to be implemented in derived classes.

        Returns
        -------
        molar_mass : array of floats
            Molar mass in [kg/mol].
        """
        return np.array([self.sublayers[i].molar_mass
                         for i in range(len(self.sublayers))])

    @material_property
    def molar_volume(self):
        """
        Returns molar volumes across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.V`.

        Returns
        -------
        molar_volume : array of floats
            Molar volumes in [m^3/mol] at the predefined radii.
        """
        return np.array([self.sublayers[i].molar_volume
                         for i in range(len(self.sublayers))])

    @material_property
    def density(self):
        """
        Returns the densities across this layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.rho`.

        Returns
        -------
        density : array of floats
            The densities of this material in [kg/m^3] at the predefined radii.
        """
        return np.array([self.sublayers[i].density
                         for i in range(len(self.sublayers))])

    @material_property
    def molar_entropy(self):
        """
        Returns molar entropies acroos the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.S`.

        Returns
        -------
        molar_entropy : array of floats
            Entropies in [J/K/mol] at the predefined radii.
        """
        return np.array([self.sublayers[i].molar_entropy
                         for i in range(len(self.sublayers))])

    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpies across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.H`.

        Returns
        -------
        molar_enthalpy : array of floats
            Enthalpies in [J/mol] at the predefined radii.
        """
        return np.array([self.sublayers[i].molar_enthalpy
                         for i in range(len(self.sublayers))])

    @material_property
    def isothermal_bulk_modulus(self):
        """
        Returns isothermal bulk moduli across the layer.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.K_T`.

        Returns
        -------
        isothermal_bulk_modulus : array of floats
            Bulk moduli in [Pa] at the predefined radii.
        """
        return np.array([self.sublayers[i].isothermal_bulk_modulus
                         for i in range(len(self.sublayers))])

    @material_property
    def adiabatic_bulk_modulus(self):
        """
        Returns the adiabatic bulk moduli across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.K_S`.

        Returns
        -------
        adiabatic_bulk_modulus : array of floats
            Adiabatic bulk modulus in [Pa] at the predefined radii.
        """
        return np.array([self.sublayers[i].adiabatic_bulk_modulus
                         for i in range(len(self.sublayers))])

    @material_property
    def isothermal_compressibility(self):
        """
        Returns isothermal compressibilities across the layer
        (or inverse isothermal bulk moduli).

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.beta_T`.

        Returns
        -------
        (K_T)^-1 : array of floats
            Compressibilities in [1/Pa] at the predefined radii.
        """
        return np.array([self.sublayers[i].isothermal_compressibility
                         for i in range(len(self.sublayers))])

    @material_property
    def adiabatic_compressibility(self):
        """
        Returns adiabatic compressibilities across the layer
        (or inverse adiabatic bulk moduli).

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.beta_S`.

        Returns
        -------
        adiabatic_compressibility : array of floats
            adiabatic compressibilities in [1/Pa] at the predefined radii.
        """
        return np.array([self.sublayers[i].adiabatic_compressibility
                         for i in range(len(self.sublayers))])

    @material_property
    def shear_modulus(self):
        """
        Returns shear moduli across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.beta_G`.

        Returns
        -------
        shear_modulus : array of floats
            Shear modulie in [Pa]  at the predefined radii.
        """
        return np.array([self.sublayers[i].shear_modulus
                         for i in range(len(self.sublayers))])

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speeds across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.v_p`.

        Returns
        -------
        p_wave_velocity : array of floats
            P wave speeds in [m/s] at the predefined radii.
        """
        return np.array([self.sublayers[i].p_wave_velocity
                         for i in range(len(self.sublayers))])

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speeds across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.v_phi`.

        Returns
        -------
        bulk sound velocity: array of floats
            Sound velocities in [m/s] at the predefined radii.
        """
        return np.array([self.sublayers[i].bulk_sound_velocity
                         for i in range(len(self.sublayers))])

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speeds across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.v_s`.

        Returns
        -------
        shear_wave_velocity : array of floats
            Wave speeds in [m/s] at the predefined radii.
        """
        return np.array([self.sublayers[i].shear_wave_velocity
                         for i in range(len(self.sublayers))])

    @material_property
    def grueneisen_parameter(self):
        """
        Returns the grueneisen parameters across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.gr`.

        Returns
        -------
        gr : array of floats
            Grueneisen parameters [unitless] at the predefined radii.
        """
        return np.array([self.sublayers[i].grueneisen_parameter
                         for i in range(len(self.sublayers))])

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficients across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.alpha`.

        Returns
        -------
        alpha : array of floats
            Thermal expansivities in [1/K] at the predefined radii.
        """
        return np.array([self.sublayers[i].thermal_expansivity
                         for i in range(len(self.sublayers))])

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volumes across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.C_v`.

        Returns
        -------
        molar_heat_capacity_v : array of floats
            Heat capacities in [J/K/mol] at the predefined radii.
        """
        return np.array([self.sublayers[i].molar_heat_capacity_v
                         for i in range(len(self.sublayers))])

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar_heat capacity at constant pressures across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.C_p`.

        Returns
        -------
        molar_heat_capacity_p : array of floats
            Heat capacities in [J/K/mol] at the predefined radii.
        """
        return np.array([self.sublayers[i].molar_heat_capacity_p
                         for i in range(len(self.sublayers))])

    # Aliased properties
    @property
    def P(self):
        """Alias for :func:`~burnman.Layer.pressure`"""
        return self.pressure

    @property
    def T(self):
        """Alias for :func:`~burnman.Layer.temperature`"""
        return self.temperature

    @property
    def energy(self):
        """Alias for :func:`~burnman.Layer.molar_internal_energy`"""
        return self.molar_internal_energy

    @property
    def helmholtz(self):
        """Alias for :func:`~burnman.Layer.molar_helmholtz`"""
        return self.molar_helmholtz

    @property
    def gibbs(self):
        """Alias for :func:`~burnman.Layer.molar_gibbs`"""
        return self.molar_gibbs

    @property
    def V(self):
        """Alias for :func:`~burnman.Layer.molar_volume`"""
        return self.molar_volume

    @property
    def rho(self):
        """Alias for :func:`~burnman.Layer.density`"""
        return self.density

    @property
    def S(self):
        """Alias for :func:`~burnman.Layer.molar_entropy`"""
        return self.molar_entropy

    @property
    def H(self):
        """Alias for :func:`~burnman.Layer.molar_enthalpy`"""
        return self.molar_enthalpy

    @property
    def K_T(self):
        """Alias for :func:`~burnman.Layer.isothermal_bulk_modulus`"""
        return self.isothermal_bulk_modulus

    @property
    def K_S(self):
        """Alias for :func:`~burnman.Layer.adiabatic_bulk_modulus`"""
        return self.adiabatic_bulk_modulus

    @property
    def beta_T(self):
        """Alias for :func:`~burnman.Layer.isothermal_compressibility`"""
        return self.isothermal_compressibility

    @property
    def beta_S(self):
        """Alias for :func:`~burnman.Layer.adiabatic_compressibility`"""
        return self.adiabatic_compressibility

    @property
    def G(self):
        """Alias for :func:`~burnman.Layer.shear_modulus`"""
        return self.shear_modulus

    @property
    def v_p(self):
        """Alias for :func:`~burnman.Layer.p_wave_velocity`"""
        return self.p_wave_velocity

    @property
    def v_phi(self):
        """Alias for :func:`~burnman.Layer.bulk_sound_velocity`"""
        return self.bulk_sound_velocity

    @property
    def v_s(self):
        """Alias for :func:`~burnman.Layer.shear_wave_velocity`"""
        return self.shear_wave_velocity

    @property
    def gr(self):
        """Alias for :func:`~burnman.Layer.grueneisen_parameter`"""
        return self.grueneisen_parameter

    @property
    def alpha(self):
        """Alias for :func:`~burnman.Layer.thermal_expansivity`"""
        return self.thermal_expansivity

    @property
    def C_v(self):
        """Alias for :func:`~burnman.Material.molar_heat_capacity_v`"""
        return self.molar_heat_capacity_v

    @property
    def C_p(self):
        """Alias for :func:`~burnman.Material.molar_heat_capacity_p`"""
        return self.molar_heat_capacity_p



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
from sympy import Matrix
from scipy.linalg import block_diag

import logging
import importlib
from ..classes.polytope import MaterialPolytope, independent_row_indices
from ..classes.solidsolution import SolidSolution
from ..classes.composite import Composite
from .solution import transform_solution_to_new_basis

try:
    cp = importlib.import_module('cvxpy')
except ImportError as err:
    print(f'Warning: {err}. '
          'For full functionality of BurnMan, please install cvxpy.')


def solution_polytope_from_charge_balance(charges, charge_total,
                                          return_fractions=False):
    """
    Creates a polytope object from a list of the charges for each species on
    each site and the total charge for all site-species.

    Parameters
    ----------
    charges : 2D list of floats
        2D list containing the total charge for species j on site i,
        including the site multiplicity. So, for example,
        a solution with the site formula [Mg,Fe]3[Mg,Al,Si]2Si3O12 would
        have the following list: [[6., 6.], [4., 6., 8.]].

    charge_total : float
        The total charge for all site-species per formula unit.
        The example given above would have charge_total = 12.

    return_fractions : boolean
        Determines whether the created polytope object returns its
        attributes (such as endmember occupancies) as fractions or as floats.
        Default is False.

    Returns
    -------
    polytope : :class:`burnman.polytope.MaterialPolytope` object
        A polytope object corresponding to the parameters provided.
    """
    n_sites = len(charges)
    all_charges = np.concatenate(charges)
    n_site_elements = len(all_charges)
    equalities = np.empty((n_sites+1, n_site_elements+1))
    equalities[:-1, 0] = -1
    i = 0
    for i_site, site_charges in enumerate(charges):
        equalities[i_site, 1:] = [1 if (j >= i and j < i+len(site_charges))
                                  else 0 for j in range(n_site_elements)]
        i += len(site_charges)

    equalities[-1, 0] = -charge_total
    equalities[-1, 1:] = all_charges

    pos_constraints = np.concatenate((np.zeros((len(equalities[0])-1, 1)),
                                      np.identity(len(equalities[0]) - 1)),
                                     axis=1)
    return MaterialPolytope(equalities, pos_constraints,
                            return_fractions=return_fractions)


def solution_polytope_from_endmember_occupancies(endmember_occupancies,
                                                 return_fractions=False):
    """
    Creates a polytope object from a list of independent endmember occupancies.

    Parameters
    ----------
    endmember_occupancies : 2D numpy array
        2D list containing the site-species occupancies j for endmember i.
        So, for example, a solution with independent endmembers
        [Mg]3[Al]2Si3O12, [Mg]3[Mg0.5Si0.5]2Si3O12, [Fe]3[Al]2Si3O12
        might have the following array:
        [[1., 0., 1., 0., 0.],
        [1., 0., 0., 0.5, 0.5],
        [0., 1., 1., 0., 0.]],
        where the order of site-species is Mg_A, Fe_A, Al_B, Mg_B, Si_B.

    return_fractions : boolean
        Determines whether the created polytope object returns its
        attributes (such as endmember occupancies) as fractions or as floats.
        Default is False.

    Returns
    -------
    polytope : :class:`burnman.polytope.MaterialPolytope` object
        A polytope object corresponding to the parameters provided.
    """
    n_sites = sum(endmember_occupancies[0])
    n_occs = endmember_occupancies.shape[1]

    nullspace = np.array(Matrix(endmember_occupancies).nullspace(),
                         dtype=float)

    equalities = np.zeros((len(nullspace)+1, n_occs+1))
    equalities[0, 0] = -n_sites
    equalities[0, 1:] = 1

    if len(nullspace) > 0:
        try:
            equalities[1:, 1:] = nullspace
        except ValueError:
            equalities[1:, 1:] = nullspace[:, :, 0]

    pos_constraints = np.concatenate((np.zeros((len(equalities[0])-1, 1)),
                                      np.identity(len(equalities[0]) - 1)),
                                     axis=1)

    return MaterialPolytope(equalities, pos_constraints,
                            return_fractions=return_fractions,
                            independent_endmember_occupancies=endmember_occupancies)


def composite_polytope_at_constrained_composition(composite, composition,
                                                  return_fractions=False):
    """
    Creates a polytope object from a Composite object and a composition.
    This polytope describes the complete set of valid composite
    endmember amounts that satisfy the compositional constraints.

    Parameters
    ----------
    composite : :class:`burnman.Composite` object
        A composite containing one or more SolidSolution and Mineral objects.

    composition : dictionary
        A dictionary containing the amounts of each element.

    return_fractions : boolean
        Determines whether the created polytope object returns its
        attributes (such as endmember occupancies) as fractions or as floats.
        Default is False.

    Returns
    -------
    polytope : :class:`burnman.polytope.MaterialPolytope` object
        A polytope object corresponding to the parameters provided.
    """
    c_array = np.empty((composite.n_elements, 1))
    c_array[:, 0] = [-composition[e] if e in composition else 0.
                     for e in composite.elements]

    equalities = np.concatenate((c_array, composite.stoichiometric_array.T),
                                axis=1)

    eoccs = []
    for i, ph in enumerate(composite.phases):
        if isinstance(ph, SolidSolution):
            eoccs.append(ph.solution_model.endmember_occupancies.T)
        else:
            eoccs.append(np.ones((1, 1)))

    eoccs = block_diag(*eoccs)
    inequalities = np.concatenate((np.zeros((len(eoccs), 1)), eoccs),
                                  axis=1)

    return MaterialPolytope(equalities, inequalities,
                            number_type='float',
                            return_fractions=return_fractions)


def simplify_composite_with_composition(composite, composition):
    """
    Takes a composite and a composition, and returns the simplest composite
    object that spans the solution space at the given composition.

    For example, if the composition is given as {'Mg': 2., 'Si': 1.5, 'O': 5.},
    and the composite is given as a mix of Mg,Fe olivine and pyroxene
    solid solutions, this function will return a composite that only contains
    the Mg-bearing endmembers.

    Parameters
    ----------
    composite : :class:`burnman.Composite` object
        The initial Composite object

    composition : dictionary
        A dictionary containing the amounts of each element

    Returns
    -------
    simple_composite : :class:`burnman.Composite` object
        The simplified Composite object
    """
    polytope = composite_polytope_at_constrained_composition(composite,
                                                             composition,
                                                             return_fractions=True)

    composite_changed = False
    new_phases = []
    mbr_amounts = polytope.endmember_occupancies
    i = 0
    for i_ph, n_mbrs in enumerate(composite.endmembers_per_phase):
        ph = composite.phases[i_ph]

        amounts = mbr_amounts[:, i:i+n_mbrs].astype(float)
        i += n_mbrs

        rank = np.linalg.matrix_rank(amounts, tol=1.e-8)

        if rank < n_mbrs:

            if isinstance(ph, SolidSolution) and rank > 0:

                if len(amounts) > 1:
                    c_mean = np.mean(amounts, axis=0)
                else:
                    c_mean = amounts[0]

                poly = solution_polytope_from_endmember_occupancies(
                    ph.solution_model.endmember_occupancies)
                dmbrs = poly.endmembers_as_independent_endmember_amounts

                x = cp.Variable(dmbrs.shape[0])
                objective = cp.Minimize(cp.sum_squares(x))
                constraints = [dmbrs.T@x == c_mean, x >= 0]

                prob = cp.Problem(objective, constraints)
                prob.solve()

                mbr_indices = np.argsort(x.value)[::-1]
                ind_indices = [i for i in mbr_indices
                               if x.value[i] > 1.e-6]
                new_basis = dmbrs[ind_indices]

                # And now reduce the new basis if necessary
                new_basis = new_basis[independent_row_indices(new_basis)]

                if len(new_basis) < ph.n_endmembers:
                    logging.info(f'Phase {i_ph} ({ph.name}) is '
                                 'rank-deficient ({rank} < {n_mbrs}). '
                                 'The transformed solution is described '
                                 f'using {len(new_basis)} endmembers.')

                    composite_changed = True
                    soln = transform_solution_to_new_basis(ph, new_basis)
                    new_phases.append(soln)
                else:
                    logging.info('This solution is rank-deficient '
                                 f'({rank} < {n_mbrs}), '
                                 'but its composition requires all '
                                 'independent endmembers.')
            else:
                composite_changed = True
                logging.info(f'Phase {i_ph} ({ph.name}) removed from '
                             'composite (rank = 0).')
        else:
            new_phases.append(ph)

    if composite_changed:
        return Composite(new_phases)
    else:
        return composite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

# TO DO: Correct heat capacity, volume where internal order-disorder is
# implemented (Landau and Bragg-Williams models)

from __future__ import absolute_import

import numpy as np

from . import equation_of_state as eos
from .. import constants

import warnings


def cork_variables(cork, cork_P, cork_T, temperature):
    a = cork[0][0] * cork_T ** (2.5) / cork_P + cork[
        0][1] * cork_T ** (1.5) / cork_P * temperature
    b = cork[1][0] * cork_T / cork_P
    c = cork[2][0] * cork_T / cork_P ** (
        1.5) + cork[2][1] / cork_P ** (1.5) * temperature
    d = cork[3][0] * cork_T / cork_P ** (
        2.0) + cork[3][1] / cork_P ** (2.0) * temperature
    return [a, b, c, d]


class CORK(eos.EquationOfState):

    """
    Class for the CoRK equation of state detailed in :cite:`HP1991`. The
    CoRK EoS is a simple virial-type extension to the modified Redlich-Kwong
    (MRK) equation of state. It was designed to compensate for the tendency of
    the MRK equation of state to overestimate volumes at high pressures and
    accommodate the volume behaviour of coexisting gas and liquid phases along
    the saturation curve.
    """

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter [unitless] as a function of pressure,
        temperature, and volume.
        """
        return 0.

    def volume(self, pressure, temperature, params):
        """
        Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
        Eq. 7 in Holland and Powell, 1991
        """
        cork = cork_variables(
            params['cork_params'], params['cork_P'], params['cork_T'], temperature)
        V = constants.gas_constant * temperature / pressure + (cork[1] - cork[0] * constants.gas_constant * np.sqrt(temperature) / (
            (constants.gas_constant * temperature + cork[1] * pressure) * (constants.gas_constant * temperature + 2. * cork[1] * pressure)) + cork[2] * np.sqrt(pressure) + cork[3] * pressure)
        return V

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].  EQ 13+2
        """
        return 0.

    # calculate the shear modulus as a function of P, V, and T
    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Not implemented.
        Returns 0.
        Could potentially apply a fixed Poissons ratio as a rough estimate.
        """
        return 0.

    # Cv, heat capacity at constant volume
    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature, and volume [J/K/mol].
        """
        return 0.

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature, and volume [1/K]
        Replace -Pth in EQ 13+1 with P-Pth for non-ambient temperature
        """
        return 0.

    # Heat capacity at ambient pressure
    def molar_heat_capacity_p0(self, temperature, params):
        """
        Returns heat capacity at ambient pressure as a function of temperature [J/K/mol]
        Cp = a + bT + cT^-2 + dT^-0.5 in Holland and Powell, 2011
        """
        Cp = params['Cp'][0] + params['Cp'][1] * temperature + params['Cp'][2] * \
            np.power(temperature, -2.) + params[
                'Cp'][3] * np.power(temperature, -0.5)
        return Cp

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure at the pressure, temperature, and volume [J/K/mol]
        """
        return 0

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        return 0.

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the gibbs free energy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        T_0 = params['T_0']
        P_relative = pressure - params['P_0']

        # Calculate temperature and pressure integrals
        intCpdT = (params['Cp'][0] * temperature + 0.5 * params['Cp'][1] * np.power(temperature, 2.) - params['Cp'][2] / temperature + 2. * params['Cp'][
                   3] * np.sqrt(temperature)) - (params['Cp'][0] * T_0 + 0.5 * params['Cp'][1] * T_0 * T_0 - params['Cp'][2] / T_0 + 2.0 * params['Cp'][3] * np.sqrt(T_0))

        intCpoverTdT = (params['Cp'][0] * np.log(temperature) + params['Cp'][1] * temperature - 0.5 * params['Cp'][2] / np.power(temperature, 2.) - 2.0 * params['Cp'][
                        3] / np.sqrt(temperature)) - (params['Cp'][0] * np.log(T_0) + params['Cp'][1] * T_0 - 0.5 * params['Cp'][2] / (T_0 * T_0) - 2.0 * params['Cp'][3] / np.sqrt(T_0))

        if params['cork_T'] == 0:
            RTlnf = 0.
        else:
            cork = cork_variables(
                params['cork_params'], params['cork_P'], params['cork_T'], temperature)

            RTlnf = constants.gas_constant * temperature * np.log(1e-5 * P_relative) + cork[1] * P_relative + cork[0] / (cork[1] * np.sqrt(temperature)) * (np.log(constants.gas_constant * temperature + cork[1] * P_relative) - np.log(
                constants.gas_constant * temperature + 2. * cork[1] * P_relative)) + 2. / 3. * cork[2] * P_relative * np.sqrt(P_relative) + cork[3] / 2. * P_relative * P_relative  # Eq. 8 in Holland and Powell, 1991

        return params['H_0'] + intCpdT - temperature * (params['S_0'] + intCpoverTdT) + RTlnf

    # calculate P = P(T0) + Pth
    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        """
        return 0.

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if 'T_0' not in params:
            params['T_0'] = 298.15
        if 'P_0' not in params:
            params['P_0'] = 0.

        # if G and Gprime are not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if 'H_0' not in params:
            params['H_0'] = float('nan')
        if 'S_0' not in params:
            params['S_0'] = float('nan')

        # check that all the required keys are in the dictionary
        expected_keys = ['cork_params', 'cork_T', 'cork_P', 'Cp']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # now check that the values are reasonable.  I mostly just
        # made up these values from experience, and we are only
        # raising a warning.  Better way to do this? [IR]

        # no test for H_0
        if params['S_0'] is not float('nan') and params['S_0'] < 0.:
            warnings.warn('Unusual value for S_0', stacklevel=2)

        if params['cork_T'] < -1.:
            warnings.warn('Unusual value for cork_T', stacklevel=2)
        if params['cork_P'] < 1.e4 or params['cork_P'] > 1.e8:
            warnings.warn('Unusual value for cork_P', stacklevel=2)

        if self.molar_heat_capacity_p0(params['T_0'], params) < 0.:
            warnings.warn('Negative heat capacity at T_0', stacklevel=2)
        if self.molar_heat_capacity_p0(2000., params) < 0.:
            warnings.warn('Negative heat capacity at 2000K', stacklevel=2)
